# -*- coding: utf-8 -*-
"""A Python API for creating workflow sessions and dynamic pipelines in *kiara*."""

import abc
import copy
import typing
import uuid
from pydantic import BaseModel, Field
from rich import box
from rich.console import Console, ConsoleOptions, RenderGroup, RenderResult
from rich.jupyter import JupyterMixin
from rich.panel import Panel
from slugify import slugify

from kiara import Kiara, KiaraModule, Pipeline, PipelineController, PipelineStructure
from kiara.data import Value, ValueSet
from kiara.info.pipelines import create_pipeline_step_table
from kiara.interfaces.python_api.controller import ApiController
from kiara.metadata.module_models import KiaraModuleInstanceMetadata
from kiara.module_config import ModuleConfig
from kiara.pipeline.config import PipelineConfig, PipelineStepConfig
from kiara.pipeline.structure import generate_pipeline_endpoint_name


class DataPointSubValue(object):
    def __init__(self, data_point: "DataPoint", sub_value_name: str):

        self._data_point: DataPoint = data_point
        self._sub_value_name: str = sub_value_name

    @property
    def point_id(self) -> str:

        return f"{self._data_point.point_id}.{self._sub_value_name}"


class DataPoint(JupyterMixin):
    def __init__(self, data_points: "DataPoints", field_name: str):

        self._points: DataPoints = data_points
        self._field_name: str = field_name
        self._links: typing.List[typing.Union[DataPoint, DataPointSubValue]] = []
        self._value: typing.Any = None
        self._sub_values: typing.Dict[str, DataPointSubValue] = {}

    def add_link(self, link: typing.Union["DataPoint", "DataPointSubValue"]):

        if self._value is not None:
            raise Exception("Can't link data point that has value set.")
        self._links.append(link)

    def set_links(self, *links: typing.Union["DataPoint", "DataPointSubValue"]):

        if self._value is not None:
            raise Exception("Can't link data point that has value set.")
        self._links.clear()
        self._links.extend(links)

    def get_links(
        self,
    ) -> typing.Iterable[typing.Union["DataPoint", "DataPointSubValue"]]:
        return self._links

    def set_value(self, value: typing.Any):

        if self._links:
            raise Exception("Can't set value for data point that has links set.")
        self._value = value

    def __getattr__(self, item):

        if item in ["value", "metadata", "data", "point_id"]:
            return self.__getattribute__(item)

        if item in self._sub_values.keys():
            return self._sub_values[item]

        self._sub_values[item] = DataPointSubValue(data_point=self, sub_value_name=item)
        return self._sub_values[item]

    @property
    def value(self) -> Value:
        step = self._points._get_step()
        if isinstance(self._points, InputDataPoints):
            value = step.workflow.get_input_value(step, self._field_name)
        else:
            value = step.workflow.get_output_value(step, self._field_name)
        value.get_metadata()
        return value

    @property
    def metadata(self) -> typing.Mapping[str, typing.Mapping[str, typing.Any]]:

        return self.value.get_metadata()

    @property
    def data(self) -> typing.Any:
        return self.value.get_value_data()

    @property
    def point_id(self) -> str:

        return f"{self._points._get_step().id}.{self._field_name}"

    def __repr__(self):
        if isinstance(self._points, InputDataPoints):
            _type = "input"
        else:
            _type = "output"
        return f"DataPoint(step={self._points._get_step().id} type={_type} field_name={self._field_name})"

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        yield self.value


class DataPoints(typing.MutableMapping[str, DataPoint]):
    def __init__(self, step: "Step"):

        self.__dict__["_data"] = {}
        self.__dict__["_step"] = step

    def __getitem__(self, item):

        return self._get_data_point(item)

    def __delitem__(self, key):

        raise NotImplementedError()

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __setitem__(self, key, value):
        raise NotImplementedError()

    def __getattr__(self, item):

        if item.startswith("_"):
            return self.__dict__[item]
        else:
            return self._get_data_point(item)

    def __setattr__(self, key, value):

        if key.startswith("_"):
            if key in self.__dict__.keys():
                raise Exception(f"{key} already set.")
            self.__dict__[key] = value
        elif key in self.__dict__["_data"].keys():
            self._update_data_point(key, self.__dict__["_data"][key], value)
        else:
            self._add_data_point(key, value)

    def _get_step(self) -> "Step":
        return self.__dict__["_step"]

    def _get_kiara(self) -> "Kiara":
        return self._get_step()._kiara

    def _get_data_point(self, field_name: str) -> DataPoint:

        if field_name in self._data.keys():
            value = self._data[field_name]
            assert isinstance(value, DataPoint)
            return self._data[field_name]
        else:
            return self._add_data_point(field_name, None)

    @abc.abstractmethod
    def _add_data_point(self, field_name: str, value: typing.Any) -> DataPoint:
        pass

    @abc.abstractmethod
    def _update_data_point(
        self, field_name: str, data_point: DataPoint, value: typing.Any
    ) -> DataPoint:
        pass


def iterable_is_all_data_points(input: typing.Iterable):

    if not isinstance(input, typing.Iterable):
        return False

    for item in input:
        if not isinstance(item, (DataPoint, DataPoints, Step, DataPointSubValue)):
            return False

    return True


class InputDataPoints(DataPoints):
    def __init__(self, step: "Step"):

        super().__init__(step=step)
        self._inputs: typing.Dict[str, typing.Any] = {}

    def _add_data_point(self, field_name: str, value: typing.Any) -> DataPoint:

        if field_name not in self._get_step().input_names:
            raise Exception(
                f"Invalid input name '{field_name}'. Available inputs: {', '.join(self._get_step().input_names)}"
            )

        dp = DataPoint(data_points=self, field_name=field_name)
        self._data[field_name] = dp
        return self._update_data_point(
            field_name=field_name, data_point=dp, value=value
        )

    def _update_data_point(
        self, field_name: str, data_point: DataPoint, value: typing.Any
    ) -> DataPoint:

        if value is None:
            # TODO: remote links/values?
            return data_point
        if isinstance(value, (DataPoint, DataPoints, Step, DataPointSubValue)):

            if isinstance(value, (DataPoint, DataPointSubValue)):
                self.add_link(value)
                data_point.set_links(value)
                return data_point
            else:
                raise NotImplementedError()

        elif isinstance(value, typing.Mapping):
            data_point.set_value(value)
            self._get_step().workflow.set_input(
                step_id=self._get_step().id, field_name=field_name, value=value
            )
            return data_point

        elif iterable_is_all_data_points(value):
            for link in value:
                self.add_link(link)

            data_point.set_links(*value)
            return data_point
        elif isinstance(value, typing.Iterable):
            data_point.set_value(value)
            self._get_step().workflow.set_input(
                step_id=self._get_step().id, field_name=field_name, value=value
            )
            return data_point
        else:
            data_point.set_value(value)
            self._get_step().workflow.set_input(
                step_id=self._get_step().id, field_name=field_name, value=value
            )
            return data_point
            # raise Exception(
            #     f"Can't update data point, invalid value type: {type(value)}"
            # )

    def add_link(
        self, link: typing.Union[DataPoint, DataPoints, "Step", DataPointSubValue]
    ):

        if isinstance(link, DataPoint):
            data_point = link
        elif isinstance(link, DataPointSubValue):
            data_point = link._data_point
        else:
            raise NotImplementedError()

        assert isinstance(data_point._points, OutputDataPoints)

        this_step: Step = self._get_step()
        other_step: Step = data_point._points._get_step()
        other_step.workflow = this_step.workflow


class OutputDataPoints(DataPoints):
    def __init__(self, step: "Step"):

        super().__init__(step=step)
        self._outputs: typing.Dict[str, typing.Any] = {}

    def _add_data_point(self, field_name: str, value: typing.Any) -> DataPoint:

        if field_name not in self._get_step().output_names:
            raise Exception(
                f"Invalid output name '{field_name}'. Available outputs: {', '.join(self._get_step().output_names)}"
            )

        dp = DataPoint(data_points=self, field_name=field_name)
        self._data[field_name] = dp
        self._update_data_point(field_name=field_name, data_point=dp, value=value)
        return dp

    def _update_data_point(
        self, field_name: str, data_point: DataPoint, value: typing.Any
    ) -> DataPoint:

        if value is not None:
            raise NotImplementedError()

        return data_point


class Workflow(JupyterMixin):
    def __init__(self, kiara: "Kiara"):

        self._kiara: Kiara = kiara
        self._id: str = str(uuid.uuid4())
        self._controller: PipelineController = ApiController()

        self._steps: typing.Dict[str, Step] = {}
        self._pipeline_conf: typing.Optional[PipelineConfig] = None
        self._structure: typing.Optional[PipelineStructure] = None
        self._pipeline: typing.Optional[Pipeline] = None

        self._inputs: typing.Dict[str, typing.Any] = {}

    @property
    def id(self) -> str:
        return self._id

    def invalidate(self):
        self._pipeline_conf = None
        self._structure = None
        self._pipeline = None

    @property
    def steps(self) -> typing.Mapping[str, "Step"]:
        return self._steps

    def add_step(self, step: "Step"):

        if step.id in self._steps.keys():
            raise Exception(f"Step '{step.id}' already part of structure.")

        self._steps[step.id] = step
        self._steps[step.id]._structure = self
        self.invalidate()

    def set_input(self, step_id: str, field_name: str, value: typing.Any):

        # TODO: validate that this is not a link?
        self._inputs.setdefault(step_id, {})[field_name] = value
        if self._pipeline is not None:
            ep_name = generate_pipeline_endpoint_name(
                step_id=step_id, value_name=field_name
            )
            self._pipeline.inputs.set_value(ep_name, value)

    def merge(self, other_workflow: "Workflow"):

        if other_workflow == self:
            return

        duplicates = set()
        other_steps = set()
        for step in other_workflow.steps.values():
            if step.id in self.steps.keys():
                duplicates.add(step.id)
            else:
                other_steps.add(step)
                other_steps.update(step.workflow.steps.values())

        if duplicates:
            raise Exception(
                f"Can't merge workflows, duplicate step id(s): {', '.join(duplicates)}"
            )

        duplicate_inputs = set()
        for k in other_workflow._inputs.keys():
            if k in self._inputs.keys():
                duplicate_inputs.add(k)

        if duplicate_inputs:
            raise Exception(
                f"Can't merge workflows, duplicate input key(s): {', '.join(duplicate_inputs)}"
            )

        for step in other_steps:
            self.add_step(step)

        self._inputs.update(other_workflow._inputs)

    @property
    def pipeline_config(self) -> PipelineConfig:

        if self._pipeline_conf is not None:
            return self._pipeline_conf

        steps: typing.List[PipelineStepConfig] = []
        for step in self.steps.values():
            input_links: typing.Dict[str, typing.List[str]] = {}
            for input_name, data_point in step.input.items():
                for link in data_point.get_links():
                    input_links.setdefault(input_name, []).append(link.point_id)

            step_conf = PipelineStepConfig(
                step_id=step.id,
                module_type=step.module_type,
                module_config=dict(step.module_config),
                input_links=input_links,  # type: ignore
            )
            steps.append(step_conf)

        self._pipeline_conf = PipelineConfig(steps=steps)
        return self._pipeline_conf

    @property
    def structure(self) -> PipelineStructure:

        if self._structure is not None:
            return self._structure

        self._structure = self.pipeline_config.create_pipeline_structure(
            kiara=self._kiara
        )
        return self._structure

    @property
    def pipeline(self) -> Pipeline:

        if self._pipeline is not None:
            return self._pipeline

        self._pipeline = Pipeline(structure=self.structure, controller=self._controller)
        inputs = {}
        for step_id, details in self._inputs.items():
            for value_name, value in details.items():
                ep_name = generate_pipeline_endpoint_name(
                    step_id=step_id, value_name=value_name
                )
                inputs[ep_name] = value
        self._pipeline.inputs.set_values(**inputs)
        return self._pipeline

    @property
    def inputs(self):

        table = self.pipeline.inputs._create_rich_table()
        return table

    @property
    def outputs(self):

        table = self.pipeline.outputs._create_rich_table()
        return table

    def get_input_value(
        self, step: typing.Union[str, "Step"], field_name: str
    ) -> Value:

        if isinstance(step, Step):
            step = step.id

        return self.pipeline.controller.get_pipeline_input(  # type: ignore
            step_id=step, field_name=field_name
        )

    def get_output_value(
        self, step: typing.Union[str, "Step"], field_name: str
    ) -> Value:

        if isinstance(step, Step):
            step = step.id

        return self.pipeline.controller.get_pipeline_output(  # type: ignore
            step_id=step, field_name=field_name
        )

    def get_all_output_values(self) -> ValueSet:
        return self.pipeline.controller.pipeline_outputs

    def process(self):

        return self.pipeline.controller.process_pipeline()

    def __eq__(self, other):

        if not isinstance(other, Workflow):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        yield self.pipeline


class Step(JupyterMixin):
    def __init__(
        self,
        module_type: str,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        step_id: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: Kiara = kiara

        self._module_type: str = module_type
        if module_config is None:
            module_config = {}
        else:
            module_config = dict(copy.deepcopy(module_config))
        self._module_config: typing.Dict[str, typing.Any] = module_config
        self._module: typing.Optional[KiaraModule] = None

        if not step_id:
            add_uuid = False
            if "module_type" in self._module_config.keys():
                if add_uuid:
                    stem = f"{self._module_type}_{self._module_config['module_type']}_{uuid.uuid4()}"
                else:
                    stem = f"{self._module_type}_{self._module_config['module_type']}"
            else:
                if add_uuid:
                    stem = f"{self._module_type}_{uuid.uuid4()}"
                else:
                    stem = self._module_type

            step_id = slugify(stem, separator="_")

        self._step_id: str = step_id

        self._inputs: InputDataPoints = InputDataPoints(self)
        self._outputs: OutputDataPoints = OutputDataPoints(self)

        self._profile_config: typing.Optional[ModuleConfig] = None

        self._structure: typing.Optional[Workflow] = None

        self._info: typing.Optional[StepInfo] = None

    @property
    def id(self) -> str:
        return self._step_id

    @property
    def info(self) -> "StepInfo":
        if self._info is None:
            module_metadata = KiaraModuleInstanceMetadata.from_module_obj(self.module)
            self._info = StepInfo(
                step_id=self._step_id, module_metadata=module_metadata
            )
        return self._info

    @property
    def module_type(self) -> str:
        return self._module_type

    @property
    def module_config(self) -> typing.Mapping[str, typing.Any]:
        return self._module_config

    def invalidate(self) -> None:
        self._profile_config = None
        self._module = None

    def config(self) -> ModuleConfig:

        if self._profile_config is not None:
            return self._profile_config

        self._profile_config = ModuleConfig(
            module_type=self._module_type, module_config=self._module_config
        )
        return self._profile_config

    @property
    def workflow(self) -> Workflow:

        if self._structure is not None:
            return self._structure

        self._structure = Workflow(kiara=self._kiara)
        self._structure.add_step(self)
        return self._structure

    @workflow.setter
    def workflow(self, structure: Workflow):

        current_structure = self.workflow
        structure.merge(current_structure)

    @property
    def module(self) -> KiaraModule:

        if self._module is not None:
            return self._module

        self._module = self._kiara.create_module(
            module_type=self._module_type, module_config=self._module_config
        )
        return self._module

    @property
    def input_names(self):
        return self.module.input_names

    @property
    def output_names(self):
        return self.module.output_names

    @property
    def input_schemas(self):
        return self.module.input_schemas

    @property
    def output_schemas(self):
        return self.module.output_schemas

    @property
    def input(self) -> DataPoints:
        return self._inputs

    @property
    def output(self) -> DataPoints:
        return self._outputs

    def __eq__(self, other):

        if not isinstance(other, Step):
            return False

        return self.id == other.id

    def __hash__(self):

        return hash(self.id)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        current_state = self.workflow.pipeline.get_current_state()
        step = current_state.structure.steps[self.id]
        step_table = create_pipeline_step_table(current_state, step)
        yield step_table


class StepInfo(JupyterMixin, BaseModel):

    step_id: str = Field(description="The step id.")
    module_metadata: KiaraModuleInstanceMetadata = Field(
        description="The module metadata."
    )

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        doc = self.module_metadata.type_metadata.documentation.create_renderable()
        metadata = self.module_metadata.create_renderable(include_doc=False)

        panel = Panel(
            RenderGroup("", Panel(doc, box=box.SIMPLE), "", metadata),
            title=f"Step info: [b]{self.step_id}[/b] (type: [i]{self.module_metadata.type_metadata.type_id}[/i])",
            title_align="left",
        )
        yield panel



# -*- coding: utf-8 -*-
import logging
import typing
from pathlib import Path
from pydantic import BaseModel, Field, validator

from kiara.module_mgmt import ModuleManager
from kiara.module_mgmt.pipelines import (
    PipelineModuleManager,
    PipelineModuleManagerConfig,
)
from kiara.module_mgmt.python_classes import (
    PythonModuleManager,
    PythonModuleManagerConfig,
)

if typing.TYPE_CHECKING:
    from kiara import Kiara, KiaraModule

log = logging.getLogger("kiara")

try:
    from typing import Literal
except Exception:
    from typing_extensions import Literal  # type: ignore


class MergedModuleManagerConfig(BaseModel):

    module_manager_type: Literal["pipeline"]
    folders: typing.List[str] = Field(
        description="A list of folders that contain pipeline descriptions.",
        default_factory=list,
    )

    @validator("folders", pre=True)
    def _validate_folders(cls, v):

        if isinstance(v, str):
            v = [v]

        assert isinstance(v, typing.Iterable)

        result = []
        for item in v:
            if isinstance(v, Path):
                item = v.as_posix()
            assert isinstance(item, str)
            result.append(item)

        return result


class MergedModuleManager(ModuleManager):
    def __init__(
        self,
        module_managers: typing.Optional[
            typing.List[
                typing.Union[PythonModuleManagerConfig, PipelineModuleManagerConfig]
            ]
        ] = None,
        extra_pipeline_folders: typing.Iterable[str] = None,
        ignore_errors: bool = False,
    ):

        self._modules: typing.Optional[typing.Dict[str, ModuleManager]] = None
        self._module_cls_cache: typing.Dict[str, typing.Type[KiaraModule]] = {}

        self._default_python_mgr: typing.Optional[PythonModuleManager] = None
        self._default_pipeline_mgr: typing.Optional[PipelineModuleManager] = None
        self._custom_pipelines_mgr: typing.Optional[PipelineModuleManager] = None

        if extra_pipeline_folders is None:
            extra_pipeline_folders = []
        self._extra_pipeline_folders: typing.Iterable[str] = extra_pipeline_folders
        self._ignore_errors: bool = ignore_errors

        if module_managers:

            raise NotImplementedError()
            # for mmc in module_managers:
            #     mm = ModuleManager.from_config(mmc)
            #     _mms.append(mm)

        self._module_mgrs: typing.Optional[typing.List[ModuleManager]] = None

    @property
    def module_managers(self) -> typing.Iterable[ModuleManager]:

        if self._module_mgrs is not None:
            return self._module_mgrs

        _mms = [
            self.default_python_module_manager,
            self.default_pipeline_module_manager,
            self.default_custom_pipelines_manager,
        ]
        self._module_mgrs = []
        self._modules = {}
        for mm in _mms:
            self.add_module_manager(mm)

        return self._module_mgrs

    @property
    def module_map(self) -> typing.MutableMapping[str, ModuleManager]:

        if self._modules is not None:
            return self._modules

        # make sure module managers are initialized before this
        # this will also initialize the _modules attribute
        self.module_managers  # noqa
        return self._modules  # type: ignore

    @property
    def default_python_module_manager(self) -> PythonModuleManager:

        if self._default_python_mgr is None:
            self._default_python_mgr = PythonModuleManager()
        return self._default_python_mgr

    @property
    def default_pipeline_module_manager(self) -> PipelineModuleManager:

        if self._default_pipeline_mgr is None:
            self._default_pipeline_mgr = PipelineModuleManager(
                folders=None, ignore_errors=self._ignore_errors
            )
        return self._default_pipeline_mgr

    @property
    def default_custom_pipelines_manager(self) -> PipelineModuleManager:

        if self._custom_pipelines_mgr is None:
            if self._extra_pipeline_folders:
                folders = {"extra": self._extra_pipeline_folders}
            else:
                folders = {}
            self._custom_pipelines_mgr = PipelineModuleManager(
                folders=folders, ignore_errors=self._ignore_errors
            )
        return self._custom_pipelines_mgr

    @property
    def available_module_types(self) -> typing.List[str]:
        """Return the names of all available modules"""
        return sorted(set(self.module_map.keys()))

    @property
    def available_non_pipeline_module_types(self) -> typing.List[str]:
        """Return the names of all available pipeline-type modules."""

        return [
            module_type
            for module_type in self.available_module_types
            if module_type != "pipeline"
            and not self.get_module_class(module_type).is_pipeline()
        ]

    @property
    def available_pipeline_module_types(self) -> typing.List[str]:
        """Return the names of all available pipeline-type modules."""

        return [
            module_type
            for module_type in self.available_module_types
            if module_type != "pipeline"
            and self.get_module_class(module_type).is_pipeline()
        ]

    def get_module_types(self) -> typing.Iterable[str]:

        return self.available_module_types

    def is_pipeline_module(self, module_type: str):

        cls = self.get_module_class(module_type=module_type)
        return cls.is_pipeline()

    def add_module_manager(self, module_manager: ModuleManager):

        if self._module_mgrs is None:
            self.module_managers  # noqa

        for module_type in module_manager.get_module_types():
            if module_type in self.module_map.keys():
                log.warning(
                    f"Duplicate module name '{module_type}'. Ignoring all but the first."
                )
                continue
            self.module_map[module_type] = module_manager

        self._module_mgrs.append(module_manager)  # type: ignore
        self._value_types = None

    def register_pipeline_description(
        self,
        data: typing.Union[Path, str, typing.Mapping[str, typing.Any]],
        module_type_name: typing.Optional[str] = None,
        namespace: typing.Optional[str] = None,
        raise_exception: bool = False,
    ) -> typing.Optional[str]:

        # making sure the module_map attribute is populated
        self.module_map  # noqa
        name = self.default_custom_pipelines_manager.register_pipeline(
            data=data, module_type_name=module_type_name, namespace=namespace
        )

        if name in self.module_map.keys():
            if raise_exception:
                raise Exception(f"Duplicate module name: {name}")
            log.warning(f"Duplicate module name '{name}'. Ignoring all but the first.")
            return None
        else:
            self.module_map[name] = self.default_custom_pipelines_manager

            return name

    def get_module_class(self, module_type: str) -> typing.Type["KiaraModule"]:

        if module_type == "pipeline":
            from kiara import PipelineModule

            return PipelineModule

        mm = self.module_map.get(module_type, None)
        if mm is None:
            raise Exception(f"No module '{module_type}' available.")

        if module_type in self._module_cls_cache.keys():
            return self._module_cls_cache[module_type]

        cls = mm.get_module_class(module_type)
        if not hasattr(cls, "_module_type_name"):
            raise Exception(
                f"Class does not have a '_module_type_name' attribute: {cls}"
            )

        assert module_type.endswith(cls._module_type_name)  # type: ignore

        if hasattr(cls, "_module_type_id") and cls._module_type_id != "pipeline" and cls._module_type_id != module_type:  # type: ignore
            raise Exception(
                f"Can't create module class '{cls}', it already has a _module_type_id attribute and it's different to the module name '{module_type}': {cls._module_type_id}"  # type: ignore
            )

        self._module_cls_cache[module_type] = cls
        setattr(cls, "_module_type_id", module_type)
        return cls

    def create_module(
        self,
        kiara: "Kiara",
        id: typing.Optional[str],
        module_type: str,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        parent_id: typing.Optional[str] = None,
    ) -> "KiaraModule":

        mm = self.module_map.get(module_type, None)
        if mm is None:
            raise Exception(
                f"No module '{module_type}' registered. Available modules: {', '.join(self.available_module_types)}"
            )

        _ = self.get_module_class(
            module_type
        )  # just to make sure the _module_type_id attribute is added

        return mm.create_module(
            id=id,
            parent_id=parent_id,
            module_type=module_type,
            module_config=module_config,
            kiara=kiara,
        )

    def find_modules_for_package(
        self,
        package_name: str,
        include_core_modules: bool = True,
        include_pipelines: bool = True,
    ) -> typing.Dict[str, typing.Type["KiaraModule"]]:

        result = {}
        for module_type in self.available_module_types:

            if module_type == "pipeline":
                continue
            module_cls = self.get_module_class(module_type)

            module_package = module_cls.get_type_metadata().context.labels.get(
                "package", None
            )
            if module_package != package_name:
                continue
            if module_cls.is_pipeline():
                if include_pipelines:
                    result[module_type] = module_cls
            else:
                if include_core_modules:
                    result[module_type] = module_cls

        return result



# -*- coding: utf-8 -*-
import abc
import typing
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult
from rich.panel import Panel
from rich.table import Table

from kiara.data.registry import DataRegistry
from kiara.data.values import Value, ValueInfo, ValueLineage, ValueSchema, ValueSlot
from kiara.defaults import INVALID_VALUE_NAMES, SpecialValue
from kiara.metadata import MetadataModel
from kiara.utils import is_debug, log_message

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.pipeline import PipelineValuesInfo


def check_valueset_valid(value_set: typing.Mapping[str, "Value"]) -> bool:

    for field_name, item in value_set.items():
        if not item.item_is_valid():
            return False
    return True


class ValueSet(typing.MutableMapping[str, "Value"]):
    def __init__(
        self,
        read_only: bool,
        title: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
    ):

        # from kiara.data.values import ValueSchema
        # schema = ValueSchema(type="any", default=None, doc="-- n/a --")
        # self._schema = schema

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: "Kiara" = kiara

        self._read_only: bool = read_only
        if title is None:
            title = "-- n/a --"
        self._title = title

    @abc.abstractmethod
    def get_all_field_names(self) -> typing.Iterable[str]:
        pass

    @abc.abstractmethod
    def _get_value_obj(self, field_name: str):
        pass

    @abc.abstractmethod
    def _set_values(
        self,
        metadata: typing.Optional[typing.Mapping[str, MetadataModel]] = None,
        lineage: typing.Optional[ValueLineage] = None,
        **values: typing.Any,
    ) -> typing.Mapping[str, typing.Union[bool, Exception]]:
        pass

    def get_value_obj(
        self,
        field_name: str,
        ensure_metadata: typing.Union[bool, typing.Iterable[str], str] = False,
    ) -> Value:

        if field_name not in list(self.get_all_field_names()):
            raise KeyError(
                f"Field '{field_name}' not available in value set. Available fields: {', '.join(self.get_all_field_names())}"
            )

        obj: Value = self._get_value_obj(field_name)

        if ensure_metadata:
            if isinstance(ensure_metadata, bool):
                obj.get_metadata()
            elif isinstance(ensure_metadata, str):
                obj.get_metadata(ensure_metadata)
            elif isinstance(ensure_metadata, typing.Iterable):
                obj.get_metadata(*ensure_metadata)
            else:
                raise ValueError(
                    f"Invalid type '{type(ensure_metadata)}' for 'ensure_metadata' argument."
                )

        return obj

    def set_value(
        self,
        key: str,
        value: typing.Any,
        metadata: typing.Optional[typing.Mapping[str, MetadataModel]] = None,
        lineage: typing.Optional[ValueLineage] = None,
    ) -> Value:

        result = self.set_values(metadata=metadata, lineage=lineage, **{key: value})
        if isinstance(result[key], Exception):
            raise result[key]  # type: ignore
        return result[key]  # type: ignore

    def set_values(
        self,
        metadata: typing.Optional[typing.Mapping[str, MetadataModel]] = None,
        lineage: typing.Optional[ValueLineage] = None,
        **values: typing.Any,
    ) -> typing.Mapping[str, typing.Union[Value, Exception]]:
        """Batch set several values.

        If metadata is provided, it is added to all values.
        """

        if self.is_read_only():
            raise Exception("Can't set values: this value set is read-only.")

        if not values:
            return {}

        invalid: typing.List[str] = []

        for k in values.keys():
            if k not in self.get_all_field_names():
                invalid.append(k)

        if invalid:
            raise ValueError(
                f"No field(s) with name(s) {', '.join(invalid)} available, valid names: {', '.join(self.get_all_field_names())}"
            )

        resolved_values = {}
        for field_name, data in values.items():
            if isinstance(data, str) and data.startswith("value:"):
                v = self._kiara.get_value(data)
                resolved_values[field_name] = v
            else:
                resolved_values[field_name] = data

        try:
            value_set_result = self._set_values(
                metadata=metadata, lineage=lineage, **resolved_values
            )
        except Exception as e:
            log_message(str(e))
            if is_debug():
                import traceback

                traceback.print_exc()
            raise e

        result: typing.Dict[str, typing.Union[Value, Exception]] = {}
        for field in values.keys():
            if isinstance(value_set_result[field], Exception):
                result[field] = value_set_result[field]  # type: ignore
            else:
                result[field] = self.get_value_obj(field)
        return result

    def is_read_only(self):
        return self._read_only

    def items_are_valid(self) -> bool:

        return check_valueset_valid(self)
        # for field_name in self.get_all_field_names():
        #     item = self.get_value_obj(field_name)
        #     if not item.item_is_valid():
        #         return False
        # return True

    def check_invalid(self) -> typing.Optional[typing.Dict[str, str]]:
        """Check whether the value set is invalid, if it is, return a description of what's wrong."""

        invalid = {}
        for field_name, item in self.items():
            if not item.item_is_valid():
                if item.value_schema.is_required():
                    if not item.is_set:
                        msg = "not set"
                    elif item.is_none:
                        msg = "no value"
                    else:
                        msg = "n/a"
                else:
                    msg = "n/a"
                invalid[field_name] = msg
        return invalid

    def get_all_value_objects(self) -> typing.Mapping[str, typing.Any]:
        return {fn: self.get_value_obj(fn) for fn in self.get_all_field_names()}

    def get_value_data_for_fields(
        self, *field_names: str, raise_exception_when_unset: bool = False
    ) -> typing.Dict[str, typing.Any]:
        """Return the data for a one or several fields of this ValueSet.

        If a value is unset, by default 'None' is returned for it. Unless 'raise_exception_when_unset' is set to 'True',
        in which case an Exception will be raised (obviously).
        """

        result: typing.Dict[str, typing.Any] = {}
        unset: typing.List[str] = []
        for k in field_names:
            v = self.get_value_obj(k)
            if not v.is_set:
                if raise_exception_when_unset:
                    unset.append(k)
                else:
                    result[k] = None
            else:
                data = v.get_value_data()
                result[k] = data

        if unset:
            raise Exception(
                f"Can't get data for fields, one or several of the requested fields are not set yet: {', '.join(unset)}."
            )

        return result

    def get_value_data(
        self, field_name: str, raise_exception_when_unset: bool = False
    ) -> typing.Any:
        return self.get_value_data_for_fields(
            field_name, raise_exception_when_unset=raise_exception_when_unset
        )[field_name]

    def get_all_value_data(
        self, raise_exception_when_unset: bool = False
    ) -> typing.Dict[str, typing.Any]:
        return self.get_value_data_for_fields(
            *self.get_all_field_names(),
            raise_exception_when_unset=raise_exception_when_unset,
        )

    def save_all(self, aliases: typing.Optional[typing.Iterable[str]] = None):

        if aliases:
            aliases = set(aliases)

        for k, v in self.items():
            field_aliases = None
            if aliases:
                field_aliases = [f"{a}__{k}" for a in aliases]
            v.save(aliases=field_aliases)

    def __getitem__(self, item: str) -> "Value":

        return self.get_value_obj(item)

    def __setitem__(self, key: str, value):

        self.set_value(key, value)

    def __delitem__(self, key: str):

        raise Exception(f"Removing items not supported: {key}")

    def __iter__(self) -> typing.Iterator[str]:
        return iter(self.get_all_field_names())

    def __len__(self):
        return len(list(self.get_all_field_names()))

    def to_details(self, ensure_metadata: bool = False) -> "PipelineValuesInfo":

        from kiara.pipeline import PipelineValueInfo, PipelineValuesInfo

        result = {}
        for name in self.get_all_field_names():
            item = self.get_value_obj(name)
            result[name] = PipelineValueInfo.from_value_obj(
                item, ensure_metadata=ensure_metadata
            )

        return PipelineValuesInfo(values=result)

    def _create_rich_table(
        self, show_headers: bool = True, ensure_metadata: bool = False
    ) -> Table:

        table = Table(box=box.SIMPLE, show_header=show_headers)
        table.add_column("Field name", style="i")
        table.add_column("Type")
        table.add_column("Description")
        table.add_column("Required")
        table.add_column("Is set")

        for k in self.get_all_field_names():
            v = self.get_value_obj(k, ensure_metadata=ensure_metadata)
            t = v.value_schema.type
            desc = v.value_schema.doc
            if not v.value_schema.is_required():
                req = "no"
            else:
                if (
                    v.value_schema.default
                    and v.value_schema.default != SpecialValue.NO_VALUE
                    and v.value_schema.default != SpecialValue.NOT_SET
                ):
                    req = "no"
                else:
                    req = "[red]yes[/red]"
            is_set = "yes" if v.item_is_valid() else "no"
            table.add_row(k, t, desc, req, is_set)

        return table

    def __repr__(self):

        title_str = ""
        if self._title:
            title_str = f" title='{self._title}'"
        return f"{self.__class__.__name__}(field_names={list(self.keys())}{title_str})"

    def __str__(self):

        return self.__repr__()

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        title = self._title
        if title:
            postfix = f" [b]{title}[/b]"
        else:
            postfix = ""

        yield Panel(
            self._create_rich_table(show_headers=True),
            box=box.ROUNDED,
            title_align="left",
            title=f"Value-Set:{postfix}",
        )


class SlottedValueSet(ValueSet):
    @classmethod
    def from_schemas(
        cls,
        schemas: typing.Mapping[str, "ValueSchema"],
        read_only: bool = True,
        check_for_sameness=True,
        initial_values: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        title: typing.Optional[str] = None,
        default_value: typing.Any = SpecialValue.NO_VALUE,
        kiara: typing.Optional["Kiara"] = None,
        registry: typing.Optional[DataRegistry] = None,
    ) -> "SlottedValueSet":

        if kiara is None:
            from kiara.kiara import Kiara

            kiara = Kiara.instance()

        if registry is None:
            registry = kiara.data_registry

        values = {}
        for field_name, schema in schemas.items():

            def_val = default_value
            if callable(default_value):
                def_val = default_value()

            _init_value = def_val

            if initial_values and initial_values.get(field_name, None) is not None:
                _init_value = initial_values[field_name]

            if not isinstance(_init_value, Value):
                value: Value = registry.register_data(
                    value_data=_init_value, value_schema=schema
                )
                # value: Value = Value(value_schema=schema, value_data=_init_value, registry=registry)  # type: ignore
            else:
                value = _init_value

            values[field_name] = value

        return cls(
            items=values,
            title=title,
            read_only=read_only,
            check_for_sameness=check_for_sameness,
            kiara=kiara,
            registry=registry,
        )

    def __init__(
        self,
        items: typing.Mapping[str, typing.Union["ValueSlot", "Value"]],
        read_only: bool,
        check_for_sameness: bool = False,
        title: typing.Optional[str] = None,
        kiara: typing.Optional["Kiara"] = None,
        registry: typing.Optional[DataRegistry] = None,
    ):
        """A `ValueSet` implementation that keeps a history of each fields value.

        Arguments:
            items: the value slots
            read_only: whether it is allowed to set new values to fields in this set
            check_for_sameness: whether a check should be performed that checks for equality of the new and old values, if equal, skip the update
            title: An optional title for this value set
            kiara: the kiara context
            registry: the registry to use to register the values to

        """

        if not items:
            raise ValueError("Can't create ValueSet: no values provided")

        self._check_for_sameness: bool = check_for_sameness

        super().__init__(read_only=read_only, title=title, kiara=kiara)

        if registry is None:
            registry = self._kiara.data_registry
        self._registry: DataRegistry = registry

        _value_slots: typing.Dict[str, ValueSlot] = {}
        for item, value in items.items():

            if value is None:
                raise Exception(
                    f"Can't create value set, item '{item}' does not have a value (yet)."
                )

            if item.startswith("_"):
                raise ValueError(f"Value name can't start with '_': {item}")
            if item in INVALID_VALUE_NAMES:
                raise ValueError(f"Invalid value name '{item}'.")

            if isinstance(value, Value):
                slot = self._registry.register_alias(value)
            elif isinstance(value, ValueSlot):
                slot = value
            else:
                raise TypeError(f"Invalid type: '{type(value)}'")

            _value_slots[item] = slot

        self._value_slots: typing.Dict[str, ValueSlot] = _value_slots

    def get_all_field_names(self) -> typing.Iterable[str]:
        return self._value_slots.keys()

    def _get_value_obj(self, field_name: str) -> "Value":

        slot = self._value_slots[field_name]
        return slot.get_latest_value()

    def _set_values(
        self,
        metadata: typing.Optional[typing.Mapping[str, MetadataModel]] = None,
        lineage: typing.Optional[ValueLineage] = None,
        **values: typing.Any,
    ) -> typing.Mapping[str, typing.Union[bool, Exception]]:

        # we want to set all registry-type values seperately, in one go, because it's more efficient
        registries: typing.Dict[
            DataRegistry, typing.Dict[ValueSlot, typing.Union["Value", typing.Any]]
        ] = {}
        field_slot_map: typing.Dict[ValueSlot, str] = {}

        result: typing.Dict[str, typing.Union[bool, Exception]] = {}
        for field_name, value_or_data in values.items():

            value_slot: ValueSlot = self._value_slots[field_name]
            if not metadata and self._check_for_sameness:
                latest_val = value_slot.get_latest_value()
                if isinstance(value_or_data, Value):
                    if latest_val.id == value_or_data.id:
                        result[field_name] = False
                        continue
                else:
                    if (
                        latest_val.is_set
                        and latest_val.get_value_data() == value_or_data
                    ):
                        result[field_name] = False
                        continue

            registries.setdefault(value_slot._registry, {})[value_slot] = value_or_data
            field_slot_map[value_slot] = field_name

        for registry, value_slots_details in registries.items():

            _r = registry.update_value_slots(value_slots_details, metadata=metadata, lineage=lineage)  # type: ignore

            for value_slot, details in _r.items():
                result[field_slot_map[value_slot]] = details

        return result


class ValuesInfo(object):
    def __init__(self, value_set: ValueSet, title: typing.Optional[str] = None):

        self._value_set: ValueSet = value_set
        self._title: typing.Optional[str] = title

    def create_value_info_table(
        self, show_headers: bool = False, ensure_metadata: bool = False
    ) -> Table:

        table = Table(show_header=show_headers, box=box.SIMPLE)
        table.add_column("Field name", style="i")
        table.add_column("Value info")

        for field_name in self._value_set.get_all_field_names():
            details: Value = self._value_set.get_value_obj(field_name)
            if not details.is_set:
                if details.item_is_valid():
                    value_info: typing.Union[
                        str, Table, ValueInfo
                    ] = "[green]-- not set --[/green]"
                else:
                    value_info = "[red]-- not set --[/red]"
            elif details.is_none:
                if details.item_is_valid():
                    value_info = "-- no value --"
                else:
                    value_info = "[red]-- no value --[/red]"
            else:
                if ensure_metadata:
                    details.get_metadata()

                value_info = details.create_info()
            table.add_row(field_name, value_info)

        return table



# -*- coding: utf-8 -*-
from kiara.data.values.value_set import ValueSet
from kiara.module import KiaraModule
from kiara.processing import JobLog, JobStatus
from kiara.processing.processor import ModuleProcessor, ProcessorConfig

try:
    pass
except Exception:
    pass


class SynchronousProcessorConfig(ProcessorConfig):

    pass


class SynchronousProcessor(ModuleProcessor):
    def process(
        self,
        job_id: str,
        module: KiaraModule,
        inputs: ValueSet,
        outputs: ValueSet,
        job_log: JobLog,
    ):

        self.job_status_updated(job_id=job_id, status=JobStatus.STARTED)
        try:
            module.process_step(inputs=inputs, outputs=outputs, job_log=job_log)
            # output_wrap._sync()
            self.job_status_updated(job_id=job_id, status=JobStatus.SUCCESS)
        except Exception as e:
            self.job_status_updated(job_id=job_id, status=e)

    def _wait_for(self, *job_ids: str):

        # jobs will always be finished, since we were waiting for them in the 'process' method
        return



# -*- coding: utf-8 -*-
"""Type-related subcommands for the cli."""
import click

from kiara import Kiara
from kiara.data.types import ValueTypesInfo
from kiara.info.types import ValueTypeInfo
from kiara.utils.output import rich_print


@click.group(name="type")
@click.pass_context
def type_group(ctx):
    """Information about available value types, and details about them."""


@type_group.command(name="list")
@click.option("--details", "-d", is_flag=True, help="Display full description.")
@click.pass_context
def list_types(ctx, details):
    """List available types (work in progress)."""

    kiara_obj: Kiara = ctx.obj["kiara"]

    print()
    value_types_info = ValueTypesInfo(kiara_obj.value_types, details=details)

    rich_print(value_types_info)


@type_group.command(name="explain")
@click.argument("value_type", nargs=1, required=True)
@click.pass_context
def explain_module_type(ctx, value_type: str):
    """Print details of a module type.

    This is different to the 'explain-instance' command, because module types need to be
    instantiated with configuration, before we can query all their properties (like
    input/output types).
    """

    kiara_obj: Kiara = ctx.obj["kiara"]

    vt_cls = kiara_obj.type_mgmt.get_value_type_cls(value_type)
    info = ValueTypeInfo.from_type_class(vt_cls, kiara=kiara_obj)

    rich_print()
    rich_print(info.create_panel(title=f"Value type: [b i]{value_type}[/b i]"))



# -*- coding: utf-8 -*-
import abc
import typing
from pydantic import Field

from kiara import Kiara, KiaraModule
from kiara.data import ValueSet
from kiara.data.values import Value, ValueSchema
from kiara.exceptions import KiaraProcessingException
from kiara.module_config import ModuleTypeConfigSchema
from kiara.operations import Operation, OperationType
from kiara.utils import log_message


class TypeConversionModuleConfig(ModuleTypeConfigSchema):

    source_type: str = Field(description="The type of the source value.")
    target_type: str = Field(description="The type of the converted value.")
    allow_none_input: bool = Field(
        description="Whether to allow 'none' source values, if one is encountered 'none' is returned.",
        default=False,
    )


class ConvertValueModule(KiaraModule):

    _config_cls = TypeConversionModuleConfig

    @classmethod
    def retrieve_module_profiles(
        cls, kiara: "Kiara"
    ) -> typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], Operation]]:

        all_metadata_profiles: typing.Dict[
            str, typing.Dict[str, typing.Dict[str, typing.Any]]
        ] = {}

        value_types: typing.Iterable[str] = cls.get_supported_value_types()
        if "*" in value_types:
            value_types = kiara.type_mgmt.value_type_names

        for value_type in value_types:

            if value_type not in kiara.type_mgmt.value_type_names:
                log_message(
                    f"Ignoring type convert operation for source type '{value_type}': type not available"
                )
                continue

            for target_type in cls.get_target_value_types():

                if target_type not in kiara.type_mgmt.value_type_names:
                    log_message(
                        f"Ignoring type convert operation for target type '{target_type}': type not available"
                    )
                    continue

                mod_conf = {
                    "source_type": value_type,
                    "target_type": target_type,
                }

                op_config = {
                    "module_type": cls._module_type_id,  # type: ignore
                    "module_config": mod_conf,
                    "doc": f"Convert value of type '{value_type}' to type '{target_type}'.",
                }
                key = f"{value_type}.convert_to.{target_type}"
                if key in all_metadata_profiles.keys():
                    raise Exception(f"Duplicate profile key: {key}")
                all_metadata_profiles[key] = op_config
                # key = f"{target_type}.convert_from.{value_type}"
                # if key in all_metadata_profiles.keys():
                #     raise Exception(f"Duplicate profile key: {key}")
                # all_metadata_profiles[key] = op_config

            for source_type in cls.get_source_value_types():

                mod_conf = {
                    "source_type": source_type,
                    "target_type": value_type,
                }

                op_config = {
                    "module_type": cls._module_type_id,  # type: ignore
                    "module_config": mod_conf,
                    "doc": f"Convert value of type '{source_type}' to type '{value_type}'.",
                }
                # key = f"{value_type}.convert_from.{target_type}"   # type: ignore
                # if key in all_metadata_profiles.keys():
                #     raise Exception(f"Duplicate profile key: {key}")
                # all_metadata_profiles[key] = op_config
                key = f"{source_type}.convert_to.{value_type}"
                if key in all_metadata_profiles.keys():
                    raise Exception(f"Duplicate profile key: {key}")
                all_metadata_profiles[key] = op_config

        return all_metadata_profiles

    @classmethod
    def get_supported_value_types(cls) -> typing.Set[str]:

        _types = cls._get_supported_value_types()
        if isinstance(_types, str):
            _types = [_types]

        return set(_types)

    @classmethod
    @abc.abstractmethod
    def _get_supported_value_types(cls) -> typing.Union[str, typing.Iterable[str]]:
        pass

    @classmethod
    def get_source_value_types(cls) -> typing.Iterable[str]:

        supported = cls.get_supported_value_types()

        types = []
        for attr_name, attr in cls.__dict__.items():

            if attr_name.startswith("from_") and callable(attr):
                v_type = attr_name[5:]
                if v_type in supported:
                    raise Exception(
                        f"Invalid configuration for type conversion module class {cls.__name__}: conversion source type can't be '{v_type}' because this is already registered as a supported type of the class."
                    )
                types.append(v_type)

        return types

    @classmethod
    def get_target_value_types(cls) -> typing.Iterable[str]:

        supported = cls.get_supported_value_types()

        types = []
        for attr_name, attr in cls.__dict__.items():

            if attr_name.startswith("to_") and callable(attr):
                v_type = attr_name[3:]
                if v_type in supported:
                    raise Exception(
                        f"Invalid configuration for type conversion module class {cls.__name__}: conversion target type can't be '{v_type}' because this is already registered as a supported type of the class."
                    )
                types.append(attr_name[3:])

        return types

    def create_input_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        supported = self.get_supported_value_types()
        source_type = self.get_config_value("source_type")
        target_type = self.get_config_value("target_type")

        if source_type in supported:
            # means this is a 'from supported type' conversion
            if target_type not in self.get_target_value_types():
                raise Exception(
                    f"Can't create input schema for module '{self._module_type_id}': converstion to target value type '{target_type}' not supported."  # type: ignore
                )

        schema: typing.Dict[str, typing.Dict[str, typing.Any]] = {
            "value_item": {
                "type": source_type,
                "doc": f"The '{source_type}' value to be converted.",
            }
        }
        if self.get_config_value("allow_none_input"):
            schema["value_item"]["optional"] = True

        return schema

    def create_output_schema(
        self,
    ) -> typing.Mapping[
        str, typing.Union[ValueSchema, typing.Mapping[str, typing.Any]]
    ]:

        supported = self.get_supported_value_types()
        source_type = self.get_config_value("source_type")
        target_type = self.get_config_value("target_type")

        if target_type in supported:
            # means this is a 'to supported type' conversion
            if source_type not in self.get_source_value_types():
                raise Exception(
                    f"Can't create output schema for module '{self._module_type_id}': conversion from source value type '{source_type}' not supported."  # type: ignore
                )

        schema: typing.Dict[str, typing.Dict[str, typing.Any]] = {
            "value_item": {
                "type": target_type,
                "doc": f"The converted '{source_type}' value as '{target_type}'.",
            }
        }
        if self.get_config_value("allow_none_input"):
            schema["value_item"]["optional"] = True

        return schema

    def process(self, inputs: ValueSet, outputs: ValueSet) -> None:

        supported_types = self.get_supported_value_types()
        source_type: str = self.get_config_value("source_type")
        target_type: str = self.get_config_value("target_type")

        allow_none: bool = self.get_config_value("allow_none_input")

        source: Value = inputs.get_value_obj("value_item")
        if source_type != source.type_name:
            raise KiaraProcessingException(
                f"Invalid type ({source.type_name}) of source value: expected '{source_type}'."
            )

        if not source.is_set or source.is_none:
            if allow_none:
                outputs.set_value("value_item", None)
                return
            else:
                raise KiaraProcessingException("No source value set.")

        if source_type in supported_types:
            # means this is a 'to' conversion

            if not hasattr(self, f"to_{target_type}"):
                # this can never happen, I think
                raise Exception(
                    f"Module '{self._module_type_id}' can't convert '{source_type}' into '{target_type}': missing method 'to_{target_type}'. This is a bug."  # type: ignore
                )

            func = getattr(self, f"to_{target_type}")
        elif target_type in supported_types:
            # menas this is a 'from' conversion
            if not hasattr(self, f"from_{source_type}"):
                # this can never happen, I think
                raise Exception(
                    f"Module '{self._module_type_id}' can't convert '{source_type}' into '{target_type}': missing method 'from_{source_type}'. This is a bug."  # type: ignore
                )

            func = getattr(self, f"from_{source_type}")

        converted = func(source)
        outputs.set_value("value_item", converted)


class ConvertValueOperationType(OperationType):
    def is_matching_operation(self, op_config: Operation) -> bool:

        return issubclass(op_config.module_cls, ConvertValueModule)

    def get_operations_for_source_type(
        self, value_type: str
    ) -> typing.Dict[str, Operation]:
        """Find all operations that transform from the specified type.

        The result dict uses the target type of the conversion as key, and the operation itself as value.
        """

        result: typing.Dict[str, Operation] = {}
        for o_id, op in self.operation_configs.items():
            source_type = op.module_config["source_type"]
            if source_type == value_type:
                target_type = op.module_config["target_type"]
                if target_type in result.keys():
                    raise Exception(
                        f"Multiple operations to transform from '{source_type}' to {target_type}"
                    )
                result[target_type] = op

        return result

    def get_operations_for_target_type(
        self, value_type: str
    ) -> typing.Dict[str, Operation]:
        """Find all operations that transform to the specified type.

        The result dict uses the source type of the conversion as key, and the operation itself as value.
        """

        result: typing.Dict[str, Operation] = {}
        for o_id, op in self.operation_configs.items():
            target_type = op.module_config["target_type"]
            if target_type == value_type:
                source_type = op.module_config["source_type"]
                if source_type in result.keys():
                    raise Exception(
                        f"Multiple operations to transform from '{source_type}' to {target_type}"
                    )
                result[source_type] = op
        return result



# -*- coding: utf-8 -*-
import typing
from pydantic import BaseModel
from rich.syntax import Syntax

from kiara.data.types import ValueType
from kiara.data.values import ValueInfo, ValueLineage
from kiara.metadata.data import DeserializeConfig, LoadConfig

if typing.TYPE_CHECKING:
    from kiara.data.values import Value


class AnyType(ValueType):
    """Any type / No type information."""

    _value_type_name = "any"

    def pretty_print_as_renderables(
        self, value: "Value", print_config: typing.Mapping[str, typing.Any]
    ) -> typing.Any:

        data = value.get_value_data()
        return [str(data)]


class KiaraInternalValueType(ValueType):
    def pretty_print_as_renderables(
        self, value: "Value", print_config: typing.Mapping[str, typing.Any]
    ) -> typing.Any:

        _value: BaseModel = value.get_value_data()
        return Syntax(_value.json(exclude_none=True, indent=2), "json")


class ValueInfoData(KiaraInternalValueType):
    """A dictionary representing a kiara ValueInfo object."""

    @classmethod
    def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
        return [typing.Mapping, ValueInfo]

    @classmethod
    def type_name(cls):
        return "value_info"

    def parse_value(self, value: typing.Any) -> typing.Any:

        if isinstance(value, typing.Mapping):
            _value = ValueInfo(**value)
            return _value

    def validate(cls, value: typing.Any) -> None:

        if not isinstance(value, ValueInfo):
            raise Exception(f"Invalid type for value info: {type(value)}.")


class ValueLineageData(KiaraInternalValueType):
    """A dictionary representing a kiara ValueLineage."""

    @classmethod
    def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
        return [typing.Mapping, ValueLineage]

    @classmethod
    def type_name(cls):
        return "value_lineage"

    def parse_value(self, value: typing.Any) -> typing.Any:

        if isinstance(value, typing.Mapping):
            _value = ValueLineage(**value)
            return _value

    def validate(cls, value: typing.Any) -> None:

        if not isinstance(value, ValueLineage):
            raise Exception(f"Invalid type for value seed: {type(value)}.")


class DeserializeConfigData(KiaraInternalValueType):
    """A dictionary representing a configuration to deserialize a value.

    This contains all inputs necessary to re-create the value.
    """

    @classmethod
    def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
        return [typing.Mapping, DeserializeConfig]

    @classmethod
    def type_name(cls):
        return "deserialize_config"

    def parse_value(self, value: typing.Any) -> typing.Any:

        if isinstance(value, typing.Mapping):
            _value = DeserializeConfig(**value)
            return _value

    def validate(cls, value: typing.Any) -> None:

        if not isinstance(value, DeserializeConfig):
            raise Exception(f"Invalid type for deserialize config: {type(value)}.")


class LoadConfigData(KiaraInternalValueType):
    """A dictionary representing load config for a *kiara* value.

    The load config schema itself can be looked up via the wrapper class  '[LoadConfig][kiara.data.persistence.LoadConfig]'.
    """

    @classmethod
    def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
        return [typing.Mapping, LoadConfig]

    @classmethod
    def type_name(cls):
        return "load_config"

    def parse_value(self, value: typing.Any) -> typing.Any:

        if isinstance(value, typing.Mapping):
            _value = LoadConfig(**value)
            return _value

    def validate(cls, value: typing.Any) -> None:

        if not isinstance(value, LoadConfig):
            raise Exception(f"Invalid type for load config: {type(value)}.")


# class ValueMetadataData(ValueType):
#     """Metadata of a value that was stored in a kiara data store."""
#
#     @classmethod
#     def candidate_python_types(cls) -> typing.Optional[typing.Iterable[typing.Type]]:
#         return [ValueInfo]
#
#     @classmethod
#     def type_name(cls):
#         return "value_metadata"
#
#     @classmethod
#     def parse_value(self, value: typing.Any) -> typing.Any:
#
#         if isinstance(value, typing.Mapping):
#             _value = ValueInfo(**value)
#             return _value
#
#     @classmethod
#     def validate(cls, value: typing.Any) -> None:
#
#         if not isinstance(value, ValueInfo):
#             raise Exception(f"Invalid type for value_metadata: {type(value)}.")



# -*- coding: utf-8 -*-
import click
import sys

from kiara.info.kiara import KiaraContext
from kiara.utils.output import rich_print


@click.command()
@click.argument("topic", nargs=1, required=False)
@click.option("--ignore-errors", "-i", help="Ignore errors.", is_flag=True)
@click.option("--json", "-j", help="Print the result in json format", is_flag=True)
@click.option(
    "--json-schema",
    "-s",
    help="Print the (json) schema of the output of this command.",
    is_flag=True,
)
@click.pass_context
def info(ctx, topic, ignore_errors, json, json_schema):
    """kiara context information"""

    kiara_obj = ctx.obj["kiara"]

    if json_schema:
        schema = KiaraContext.schema_json(indent=2)
        # s = Syntax(schema, "json", background_color="default")  # doesn't seem to work with pipes
        print(schema)
        sys.exit(0)

    if not topic:
        info = KiaraContext.create(kiara=kiara_obj, ignore_errors=ignore_errors)
    else:

        if topic not in KiaraContext.__fields__.keys():
            print(
                f"Info topic '{topic}' not available. Available topics: {', '.join(KiaraContext.__fields__.keys())}"
            )
            sys.exit(1)

        _info = KiaraContext.create(kiara=kiara_obj, ignore_errors=ignore_errors)
        info = getattr(_info, topic)

    if json:
        print(info.json(indent=2))
    else:
        rich_print(info)



# -*- coding: utf-8 -*-
import typing

from kiara.data import Value
from kiara.metadata import MetadataModel
from kiara.operations import Operation
from kiara.utils.class_loading import find_all_metadata_schemas

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara
    from kiara.operations.extract_metadata import (
        ExtractMetadataModule,
        ExtractMetadataOperationType,
    )


class MetadataMgmt(object):
    def __init__(self, kiara: "Kiara"):

        self._kiara: "Kiara" = kiara
        self._metadata_models: typing.Optional[
            typing.Dict[str, typing.Type[MetadataModel]]
        ] = None

    @property
    def all_schemas(self) -> typing.Mapping[str, typing.Type[MetadataModel]]:

        if self._metadata_models is None:
            self._metadata_models = find_all_metadata_schemas()

        return self._metadata_models

    def find_all_schemas_for_package(
        self, package_name: str
    ) -> typing.Dict[str, typing.Type[MetadataModel]]:

        result = {}
        for name, schema in self.all_schemas.items():
            schema_md = schema.get_type_metadata()
            package = schema_md.context.labels.get("package")
            if package == package_name:
                result[name] = schema

        return result

    def get_metadata_keys_for_type(self, value_type: str) -> typing.Set[str]:

        all_profiles_for_type: typing.Mapping[
            str, Operation
        ] = self.get_metadata_operations_for_value_type(value_type=value_type)

        if not all_profiles_for_type:
            return set()
        else:
            return set(all_profiles_for_type.keys())

    def get_metadata_operations_for_value_type(
        self, value_type: str
    ) -> typing.Mapping[str, Operation]:

        metadata_operations: ExtractMetadataOperationType = self._kiara.operation_mgmt.get_operations("extract_metadata")  # type: ignore

        all_profiles_for_type: typing.Mapping[
            str, Operation
        ] = metadata_operations.get_all_operations_for_type(value_type)

        return all_profiles_for_type

    def get_value_metadata(
        self, value: Value, *metadata_keys: str, also_return_schema: bool = False
    ):

        if value.is_none:
            return {
                k: {"metadata_item": {}, "metadata_schema": ""} for k in metadata_keys
            }

        value_type = value.value_schema.type
        # TODO: validate type exists

        all_metadata_keys_for_type = self.get_metadata_keys_for_type(
            value_type=value_type
        )
        # all_profiles_for_type = self._kiara.operations.type_operations.extract_metadata_profiles.get(
        #     value_type, None
        # )
        # if all_profiles_for_type is None:
        #     all_profiles_for_type = {}

        if not metadata_keys:
            _metadata_keys = set(all_metadata_keys_for_type)
            # add existing, externally added metadata
            for key in value.metadata.keys():
                _metadata_keys.add(key)
        elif isinstance(metadata_keys, str):
            _metadata_keys = {metadata_keys}
        elif isinstance(metadata_keys, typing.Iterable):
            _metadata_keys = set(metadata_keys)
        else:
            raise TypeError(f"Invalid type for metadata keys: {type(metadata_keys)}")

        result = {}
        missing = []

        for md_key in all_metadata_keys_for_type:
            if md_key not in value.metadata.keys():
                missing.append(md_key)
            else:
                result[md_key] = value.metadata[md_key]

        extract_metadata_ops: ExtractMetadataOperationType = self._kiara.operation_mgmt.get_operations(  # type: ignore
            "extract_metadata"
        )
        value_md_ops = extract_metadata_ops.get_all_operations_for_type(
            value_type=value_type
        )

        for mk in missing:
            if mk not in all_metadata_keys_for_type:
                raise Exception(
                    f"Can't extract metadata profile '{mk}' for type '{value_type}': metadata profile does not exist (for this type, anyway)."
                )

            op_config = value_md_ops.get(mk, None)
            if op_config is None:
                raise Exception(
                    f"Can't extract metadata profile '{mk}' for type '{value_type}': metadata profile does not exist (for this type, anyway). This is a bug."
                )

            md_module: ExtractMetadataModule = op_config.module  # type: ignore
            md_result = md_module.run(_attach_lineage=False, value_item=value)
            result[mk] = md_result.get_all_value_data()

        if also_return_schema:
            return result
        else:
            return {k: v["metadata_item"] for k, v in result.items()}



# -*- coding: utf-8 -*-
import typing
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult
from rich.table import Table
from slugify import slugify

from kiara.data.values.value_set import ValueSet
from kiara.defaults import DEFAULT_NO_DESC_VALUE
from kiara.info.pipelines import PipelineState, StepsInfo
from kiara.module_config import ModuleConfig
from kiara.pipeline import StepStatus
from kiara.pipeline.module import PipelineModule
from kiara.pipeline.pipeline import Pipeline, PipelineController, PipelineStructure

if typing.TYPE_CHECKING:
    from kiara.kiara import Kiara


class KiaraWorkflow(object):
    """A thin wrapper class around a [PipelineModule][kiara.pipeline.PipelineModule], mostly handling initialization from simplified configuration data."""

    def __init__(
        self,
        workflow_id: str,
        config: ModuleConfig,
        kiara: "Kiara",
        controller: typing.Optional[PipelineController] = None,
    ):

        self._controller: typing.Optional[PipelineController] = controller
        self._workflow_id: str = workflow_id
        self._workflow_config: ModuleConfig = config
        self._kiara: Kiara = kiara

        root_module_args: typing.Dict[str, typing.Any] = {"id": self._workflow_id}
        if self._workflow_config.module_type == "pipeline":
            root_module_args["module_type"] = "pipeline"
            root_module_args["module_config"] = self._workflow_config.module_config
        elif self._kiara.is_pipeline_module(self._workflow_config.module_type):
            root_module_args["module_type"] = self._workflow_config.module_type
            root_module_args["module_config"] = self._workflow_config.module_config
        else:
            # means it's a python module, and we wrap it into a single-module pipeline
            root_module_args["module_type"] = "pipeline"
            steps_conf = {
                "steps": [
                    {
                        "module_type": self._workflow_config.module_type,
                        "step_id": slugify(
                            self._workflow_config.module_type, separator="_"
                        ),
                        "module_config": self._workflow_config.module_config,
                    }
                ],
                "input_aliases": "auto",
                "output_aliases": "auto",
            }
            root_module_args["module_config"] = steps_conf

        self._root_module: PipelineModule = self._kiara.create_module(**root_module_args)  # type: ignore
        assert isinstance(self._root_module, PipelineModule)
        self._pipeline: typing.Optional[Pipeline] = None

    @property
    def structure(self) -> PipelineStructure:
        return self._root_module.structure

    @property
    def pipeline(self) -> Pipeline:

        if self._pipeline is None:
            self._pipeline = Pipeline(self.structure, controller=self._controller)
        return self._pipeline

    @property
    def controller(self) -> PipelineController:
        return self.pipeline.controller

    @property
    def status(self) -> StepStatus:
        return self.pipeline.status

    @property
    def inputs(self) -> ValueSet:
        return self.pipeline.inputs

    @property
    def outputs(self) -> ValueSet:
        return self.pipeline.outputs

    def get_current_state(self) -> PipelineState:
        return self.pipeline.get_current_state()

    @property
    def current_state(self) -> PipelineState:
        return self.get_current_state()

    # @inputs.setter
    # def inputs(self, inputs: typing.Mapping[str, typing.Any]):
    #     self.pipeline.set_pipeline_inputs(**inputs)

    @property
    def input_names(self) -> typing.List[str]:
        return list(self.inputs.get_all_field_names())

    @property
    def output_names(self) -> typing.List[str]:
        return list(self.outputs.get_all_field_names())

    @property
    def workflow_id(self) -> str:
        return self._workflow_id

    @property
    def steps(self) -> StepsInfo:
        return self.pipeline.structure.to_details().steps_info

    def __repr__(self):

        return f"{self.__class__.__name__}(workflow_id={self.workflow_id}, root_module={self._root_module})"

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        yield f"[b]Workflow: {self.workflow_id}[/b]"

        doc = self._root_module.get_type_metadata().documentation.description
        if doc and doc != DEFAULT_NO_DESC_VALUE:
            yield f"\n{doc}\n"

        table = Table(box=box.SIMPLE, show_header=False)
        table.add_column("property", style="i")
        table.add_column("value")
        doc_link = self._root_module.get_type_metadata().context.references.get(
            "documentation", None
        )
        if doc_link:
            # TODO: use direct link
            url = doc_link.url
            module_str = f"[link={url}]{self._root_module._module_type_id}[/link]"  # type: ignore
        else:
            module_str = self._root_module._module_type_id  # type: ignore
        table.add_row("root module", module_str)
        table.add_row("current status", self.status.name)
        inputs_table = self.inputs._create_rich_table(show_headers=True)
        table.add_row("inputs", inputs_table)
        outputs_table = self.outputs._create_rich_table(show_headers=True)
        table.add_row("outputs", outputs_table)

        yield table



# -*- coding: utf-8 -*-

from click.testing import CliRunner

from kiara.interfaces.cli import cli


def test_run_without_module():

    runner = CliRunner()
    result = runner.invoke(cli, "run")
    assert result.exit_code == 2
    assert "Missing argument" in result.stdout


def test_run_without_args():

    runner = CliRunner()
    result = runner.invoke(cli, "run logic.and")
    print(result.stdout)
    assert result.exit_code == 0
    assert "No inputs provided" in result.stdout


def test_run_with_missing_arg():

    runner = CliRunner()
    result = runner.invoke(cli, "run logic.and a=true")
    assert result.exit_code == 3
    assert "inputs not ready yet" in result.stdout


def test_run_with_valid_inputs():

    runner = CliRunner()
    result = runner.invoke(cli, "run logic.and a=true b=true")
    assert result.exit_code == 0
    assert "True" in result.stdout


# def test_run_with_save():
#
#     runner = CliRunner()
#     result = runner.invoke(cli, "run logic.and a=true b=true")
#     assert result.exit_code == 0
#     assert "True" in result.stdout



# -*- coding: utf-8 -*-
import typing
from pydantic import Field
from rich import box
from rich.console import RenderableType
from rich.table import Table

from kiara.info import KiaraInfoModel
from kiara.info.modules import ModuleTypesGroupInfo
from kiara.info.operations import OperationsGroupInfo
from kiara.info.pipelines import PipelineTypesGroupInfo

if typing.TYPE_CHECKING:
    from kiara import Kiara


class ModulesGroup(KiaraInfoModel):
    @classmethod
    def create(cls, kiara: "Kiara", ignore_errors: bool = False):

        module_types = ModuleTypesGroupInfo.from_type_names(
            kiara=kiara, ignore_pipeline_modules=True
        )
        pipeline_types = PipelineTypesGroupInfo.create(
            kiara=kiara, ignore_errors=ignore_errors
        )

        return ModulesGroup(module_types=module_types, pipelines=pipeline_types)

    module_types: ModuleTypesGroupInfo = Field(
        description="The available module types."
    )
    pipelines: PipelineTypesGroupInfo = Field(description="The available pipelines.")

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        mod_table = self.module_types.create_renderable()
        pipe_table = self.pipelines.create_renderable()

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Type", style="b")
        table.add_column("Modules")

        table.add_row("Module types", mod_table)
        table.add_row("Pipelines", pipe_table)
        return table


class KiaraContext(KiaraInfoModel):
    @classmethod
    def create(cls, kiara: "Kiara", ignore_errors: bool = False):

        modules = ModulesGroup.create(kiara=kiara, ignore_errors=ignore_errors)
        op_group = OperationsGroupInfo.create(kiara=kiara, ignore_errors=ignore_errors)
        return KiaraContext(
            modules=modules,
            operations=op_group,
        )

    modules: ModulesGroup = Field(
        description="Information about the available modules."
    )
    operations: OperationsGroupInfo = Field(
        description="Information about supported operations in this kiara context."
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")

        module_table = self.modules.create_renderable()
        table.add_row("module types", module_table)

        operations_table = self.operations.create_renderable()
        table.add_row("operations", operations_table)

        return table



# -*- coding: utf-8 -*-
import typing

if typing.TYPE_CHECKING:
    pass

# class WrappedStep(object):
#
#     def __init__(self, workflow: "WrappedWorkflow", module_type: str, module_config: typing.Optional[typing.Mapping[str, typing.Any]]=None, input_links: typing.Optional[typing.Mapping[str, typing.Any]]=None, step_id: typing.Optional[str]=None):
#
#         self._workflow: WrappedWorkflow = workflow
#         self._module_type: str = module_type
#         if module_config is None:
#             module_config = {}
#         else:
#             module_config = dict(module_config)
#         self._module_config: typing.Dict[str, typing] = module_config
#
#         if step_id is None:
#             step_id = slugify(module_type)
#         self._step_id: str = step_id
#
#         self._step_config: typing.Optional[PipelineStepConfig] = None
#         self._module: typing.Optional[KiaraModule] = None
#
#         self._input_links: typing.Dict[str, typing.List[StepValueAddress]] = {}
#         if input_links:
#             for k, v in input_links.items():
#                 self._add_input_link(k, v)
#
#         self._input_points = DataPoints(step=self)
#         self._output_points = DataPoints(step=self)
#
#     def invalidate(self):
#
#         self._step_config = None
#         self._module = None
#         self._workflow.invalidate()
#
#     @property
#     def step_id(self) -> str:
#         return self._step_id
#
#     @property
#     def module_type(self) -> str:
#         return self._module_type
#
#     @property
#     def module_config(self) -> typing.Dict[str, typing.Mapping[str, typing.Any]]:
#         return self._module_config
#
#     @property
#     def input(self) -> DataPoints:
#         return self._input_points
#
#     @property
#     def output(self) -> DataPoint:
#         return self._output_points
#
#     @property
#     def config(self) -> PipelineStepConfig:
#
#         if self._step_config is not None:
#             return self._step_config
#
#         self._step_config = PipelineStepConfig(step_id=self.step_id, module_type=self.module_type, module_config=self.module_config, input_links=self._input_links)
#         return self._step_config
#
#     @property
#     def module(self) -> KiaraModule:
#         if self._module is not None:
#             return self._module
#
#         self._module = self.config.create_module(kiara=self._workflow.kiara)
#         return self._module
#
#     @property
#     def input_names(self) -> typing.Iterable[str]:
#         return self.module.input_names
#
#     @property
#     def output_names(self) -> typing.Iterable[str]:
#         return self.module.output_names
#
#     @property
#     def input_schemas(self) -> typing.Mapping[str, ValueSchema]:
#         return self.module.input_schemas
#
#     @property
#     def output_schemas(self) -> typing.Mapping[str, ValueSchema]:
#         return self.module.output_schemas
#
#     def connect_input(self, input_name: str, target_step: typing.Union["WrappedStep", str], sub_target: typing.Union[str, typing.Mapping, typing.Iterable]=None):
#
#         if isinstance(target_step, WrappedStep):
#             step_id = target_step.step_id
#             target_step_obj = target_step
#             output_name = None
#             _sub_target = None
#         elif isinstance(target_step, str):
#             if "." not in target_step:
#                 step_id = target_step
#                 output_name = None
#                 _sub_target = None
#             else:
#                 step_id, rest = target_step.split(".", maxsplit=1)
#                 if "." not in rest:
#                     output_name = rest
#                     _sub_target = None
#                 else:
#                     output_name, _sub_target = rest.split(".", maxsplit=1)
#             target_step_obj = self._workflow.get_step(step_id=step_id)
#
#         if sub_target is not None:
#             if _sub_target:
#                 raise Exception(f"Sub-target specified twice: {sub_target} -- {_sub_target}")
#
#             if isinstance(sub_target, str):
#                 _sub_target = sub_target
#             elif isinstance(sub_target, typing.Mapping):
#                 raise NotImplementedError()
#             elif isinstance(sub_target, typing.Iterable):
#                 raise NotImplementedError()
#             else:
#                 raise TypeError(f"Invalid type '{type(sub_target)}' for sub_target value.")
#
#         if not output_name:
#             # here we try to find the right connnecting point automatically
#             if target_step_obj is None:
#                 raise NotImplementedError()
#
#             if input_name in target_step_obj.output_schemas.keys():
#                 req_type = self.input_schemas[input_name].type
#                 avail_type = target_step_obj.output_schemas[input_name].type
#                 if req_type == avail_type:
#                     output_name = input_name
#                 else:
#                     raise NotImplementedError()
#             else:
#                 # look if the connected step has a single output of the right type
#                 req_type = self.input_schemas[input_name].type
#                 match = []
#
#                 print(req_type)
#                 for n, s in target_step_obj.output_schemas.items():
#                     print(n)
#                     print(s.type)
#                     if s.type == req_type:
#                         match.append(n)
#
#                 if len(match) == 1:
#                     output_name = match[0]
#                 else:
#                     raise NotImplementedError()
#
#         target_dict = {
#             "step_id": step_id,
#             "output_name": output_name,
#             "sub_value": _sub_target
#
#         }
#
#         self._add_input_link(input_name=input_name, target=target_dict)
#
#     def _add_input_link(self, input_name: str, target: typing.Mapping):
#
#         if input_name in self._input_links.keys():
#             raise NotImplementedError()
#         vas = ensure_step_value_addresses(default_field_name=input_name, link=target)
#         self._input_links[input_name] = vas
#         self.invalidate()
#         self._workflow.invalidate()


# class WrappedWorkflow(object):
#
#     def __init__(self, kiara: "Kiara", id: str):
#
#         self._kiara: Kiara = kiara
#         self._id = id
#
#         self._steps: typing.Dict[str, PipelineStepConfig] = {}
#         self._current_structure: typing.Optional[PipelineStructure] = None
#
#     @property
#     def id(self) -> str:
#         return self._id
#
#     @property
#     def structure(self) -> PipelineStructure:
#
#         if self._current_structure is not None:
#             return self._current_structure
#
#         new_config = {}
#         steps = [x.config for x in self._steps.values()]
#         new_config["steps"] = steps
#         config = PipelineConfig(**new_config)
#         self._current_structure = PipelineStructure(parent_id=self.id, config=config, kiara=self._kiara)
#         return self._current_structure
#
#     @property
#     def kiara(self) -> "Kiara":
#         return self._kiara
#
#     def invalidate(self):
#
#         self._current_structure = None
#
#     def add_step(self, module_type: str, module_config: typing.Optional[typing.Mapping[str, typing.Any]]=None, input_links: typing.Mapping[str, typing.Any]=None, step_id: typing.Optional[str]=None) -> WrappedStep:
#
#         step = WrappedStep(workflow=self, module_type=module_type, module_config=module_config, input_links=input_links, step_id=step_id)
#         if step.step_id in self._steps:
#             raise Exception(f"Duplicate step id: {step.step_id}")
#
#         self._steps[step.step_id] = step
#         self.invalidate()
#         return step
#
#     def get_step(self, step_id: str) -> WrappedStep:
#
#         if step_id in self._steps.keys():
#             return self._steps[step_id]
#         else:
#             raise Exception(f"No step with id: {step_id}")



# -*- coding: utf-8 -*-

from click.testing import CliRunner

from kiara.interfaces.cli import cli


def test_module_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "module")
    assert result.exit_code == 0
    assert "Module-related sub-commands" in result.stdout


def test_logic_list_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "module list")

    assert result.exit_code == 0
    assert "logic.xor" in result.stdout
    assert "table.import_from.file_path.string" not in result.stdout


def test_module_explain_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "module explain logic.xor")

    assert result.exit_code == 0
    assert "Pipeline config" in result.stdout
    assert "Processing source code" not in result.stdout
    # assert "https://github.com/DHARPA-Project/kiara_modules.core" in result.stdout

    result = runner.invoke(cli, "module explain logic.and")

    assert result.exit_code == 0
    assert "Pipeline config" not in result.stdout
    assert "Processing source code" in result.stdout
    # assert "https://github.com/DHARPA-Project/kiara_modules.core" in result.stdout


def test_module_explain_instance_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "module explain-instance logic.xor")

    assert result.exit_code == 0
    assert "Inputs" in result.stdout
    assert "constants" in result.stdout

    result = runner.invoke(cli, "module explain-instance logic.and")
    assert result.exit_code == 0
    assert "Inputs" in result.stdout
    assert "constants" in result.stdout



# -*- coding: utf-8 -*-
import threading
import typing
import uvicorn
from fastapi import APIRouter, FastAPI
from pydantic.fields import Field
from pydantic.main import BaseModel

from kiara import Kiara
from kiara.data.values import ValueSchema
from kiara.metadata.module_models import KiaraModuleInstanceMetadata


class ModuleRunResponse(BaseModel):

    outputs: typing.Dict[str, typing.Any] = Field(description="The module output.")
    output_schema: typing.Dict[str, ValueSchema] = Field(
        description="The output schema."
    )


class KiaraRestService(object):
    def __init__(self, kiara: typing.Optional[Kiara] = None):

        if kiara is None:
            kiara = Kiara.instance()

        self._kiara: Kiara = kiara
        self._app = FastAPI()
        self._modules_router: APIRouter = self.create_modules_router()
        self._app.include_router(self._modules_router, prefix="/module")

    def start(self):

        thread = threading.Thread(target=uvicorn.run, args=(self._app,))
        thread.start()
        thread.join()

    def create_modules_router(self):

        modules_router = APIRouter()

        @modules_router.post("/explain-instance")
        def explain(
            module_type: str,
            module_config: typing.Optional[typing.Dict[str, typing.Any]] = None,
        ) -> KiaraModuleInstanceMetadata:

            module = self._kiara.create_module(
                module_type=module_type, module_config=module_config
            )
            return module.info

        @modules_router.post("/run")
        def run(
            module_type: str,
            module_config: typing.Optional[typing.Dict[str, typing.Any]] = None,
            inputs: typing.Optional[typing.Dict[str, typing.Any]] = None,
        ) -> ModuleRunResponse:

            raise NotImplementedError()

            # module = self._kiara.create_module(
            #     module_type=module_type, module_config=module_config
            # )
            #
            # result_values = self._kiara.run_module(
            #     module=module,
            #     inputs=inputs,
            #     resolve_result=False,
            # )
            #
            # op_mgmt: OperationMgmt = self._kiara.operation_mgmt
            #
            # result = {}
            # for field_name, value in result_values.items():
            #     v = op_mgmt.run("serialize", "msgpack", value)
            #     result[field_name] = v.get_value_obj("bytes")
            #
            # r = {}
            # schemas = {}
            #
            # for field_name, value in result.items():
            #     assert value.type_name == "bytes"
            #     v = op_mgmt.run("deserialize", "msgpack", value)
            #     r[field_name] = v.get_value_obj("value_data")
            #     schemas[field_name] = ValueSchema(
            #         type=v.get_value_data("value_type"),
            #         doc=result_values.get_value_obj(field_name).value_schema.doc,
            #     )
            #
            # value_set = ValueSetImpl.from_schemas(
            #     kiara=self._kiara, schemas=schemas, initial_values=r
            # )
            # outputs = value_set.get_all_value_data()
            #
            # _result = ModuleRunResponse(outputs=outputs, output_schema=schemas)
            # return _result

        return modules_router



# -*- coding: utf-8 -*-
import copy
import networkx as nx
import typing
import uuid
from functools import lru_cache
from networkx import NetworkXNoPath, NodeNotFound
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator
from rich.console import Console, ConsoleOptions, RenderResult

from kiara.data.values import ValueSchema
from kiara.pipeline import StepValueAddress
from kiara.pipeline.utils import extend_pipeline, generate_step_alias
from kiara.pipeline.values import (
    PipelineInputRef,
    PipelineOutputRef,
    StepInputRef,
    StepOutputRef,
)

if typing.TYPE_CHECKING:
    from kiara.info.pipelines import PipelineStructureDesc
    from kiara.kiara import Kiara
    from kiara.module import KiaraModule
    from kiara.pipeline.config import PipelineConfig, PipelineStepConfig
    from kiara.pipeline.pipeline import Pipeline


class PipelineStep(BaseModel):
    """A step within a pipeline-structure, includes information about it's connection(s) and other metadata."""

    class Config:
        validate_assignment = True
        extra = Extra.forbid

    @classmethod
    def create_steps(
        cls, *steps: "PipelineStepConfig", kiara: "Kiara"
    ) -> typing.List["PipelineStep"]:

        result: typing.List[PipelineStep] = []
        if kiara is None:
            from kiara.module import Kiara

            kiara = Kiara.instance()

        for step in steps:

            _s = PipelineStep(
                step_id=step.step_id,
                module_type=step.module_type,
                module_config=copy.deepcopy(step.module_config),
                input_links=copy.deepcopy(step.input_links),
                _kiara=kiara,
            )
            result.append(_s)

        return result

    _module: typing.Optional["KiaraModule"] = PrivateAttr(default=None)

    @validator("step_id")
    def _validate_step_id(cls, v):

        assert isinstance(v, str)
        if "." in v:
            raise ValueError("Step ids can't contain '.' characters.")

        return v

    step_id: str
    module_type: str = Field(description="The module type.")
    module_config: typing.Mapping[str, typing.Any] = Field(
        description="The module config.", default_factory=dict
    )
    required: bool = Field(
        description="Whether this step is required within the workflow.\n\nIn some cases, when none of the pipeline outputs have a required input that connects to a step, then it is not necessary for this step to have been executed, even if it is placed before a step in the execution hierarchy. This also means that the pipeline inputs that are connected to this step might not be required.",
        default=True,
    )
    processing_stage: typing.Optional[int] = Field(
        default=None,
        description="The stage number this step is executed within the pipeline.",
    )
    input_links: typing.Mapping[str, typing.List[StepValueAddress]] = Field(
        description="The links that connect to inputs of the module.",
        default_factory=list,
    )
    _kiara: typing.Optional["Kiara"] = PrivateAttr(default=None)
    _id: str = PrivateAttr()

    def __init__(self, **data):  # type: ignore

        self._id = str(uuid.uuid4())
        kiara = data.pop("_kiara", None)
        if kiara is None:
            from kiara import Kiara

            kiara = Kiara.instance()

        super().__init__(**data)
        self._kiara: "Kiara" = kiara

    @property
    def kiara(self):
        return self._kiara

    @property
    def module(self) -> "KiaraModule":

        if self._module is None:

            try:
                if (
                    self.module_type in self.kiara.operation_mgmt.profiles.keys()
                    and not self.module_config
                ):
                    op = self.kiara.operation_mgmt.profiles[self.module_type]
                    self._module = op.module
                else:
                    self._module = self.kiara.create_module(
                        id=self.step_id,
                        module_type=self.module_type,
                        module_config=self.module_config,
                    )
            except Exception as e:
                raise Exception(f"Can't assemble pipeline structure: {e}")

        return self._module

    def __eq__(self, other):

        if not isinstance(other, PipelineStep):
            return False

        return self._id == other._id

        # # TODO: also check whether _kiara obj is equal?
        # eq = (self.step_id, self.parent_id, self.module, self.processing_stage,) == (
        #     other.step_id,
        #     other.parent_id,
        #     other.module,
        #     other.processing_stage,
        # )
        #
        # if not eq:
        #     return False
        #
        # hs = DeepHash(self.input_links)
        # ho = DeepHash(other.input_links)
        #
        # return hs[self.input_links] == ho[other.input_links]

    def __hash__(self):

        return hash(self._id)

        # # TODO: also include _kiara obj?
        # # TODO: figure out whether that can be made to work without deephash
        # hs = DeepHash(self.input_links)
        # return hash(
        #     (
        #         self.step_id,
        #         self.parent_id,
        #         self.module,
        #         self.processing_stage,
        #         hs[self.input_links],
        #     )
        # )

    def __repr__(self):

        return f"{self.__class__.__name__}(step_id={self.step_id} module_type={self.module_type} processing_stage={self.processing_stage})"

    def __str__(self):
        return f"step: {self.step_id} (module: {self.module_type})"


def generate_pipeline_endpoint_name(step_id: str, value_name: str):

    return f"{step_id}__{value_name}"


ALLOWED_INPUT_ALIAS_MARKERS = ["auto", "auto_all_outputs"]


def calculate_shortest_field_aliases(
    steps: typing.List[PipelineStep], alias_type: str, alias_for: str
):

    assert alias_for in ["inputs", "outputs"]
    if alias_type == "auto_all_outputs":

        aliases: typing.Dict[str, typing.List[str]] = {}

        for step in steps:

            if alias_for == "inputs":
                field_names = step.module.input_names
            else:
                field_names = step.module.output_names

            for field_name in field_names:
                aliases.setdefault(field_name, []).append(step.step_id)

        result = {}
        for field_name, step_ids in aliases.items():
            if len(step_ids) == 1:
                result[
                    generate_pipeline_endpoint_name(step_ids[0], field_name)
                ] = field_name
            else:
                for step_id in step_ids:
                    generated = generate_pipeline_endpoint_name(step_id, field_name)
                    result[generated] = generated

    elif alias_type == "auto":

        aliases = {}

        for stage_nr, step in enumerate(steps):

            _field_names: typing.Optional[typing.Iterable[str]] = None
            if alias_for == "inputs":
                _field_names = step.module.input_names
            else:
                if stage_nr == len(steps) - 1:
                    _field_names = step.module.output_names

            if not _field_names:
                continue

            for field_name in _field_names:
                aliases.setdefault(field_name, []).append(step.step_id)

        result = {}
        for field_name, step_ids in aliases.items():
            if len(step_ids) == 1:
                result[
                    generate_pipeline_endpoint_name(step_ids[0], field_name)
                ] = field_name
            else:
                for step_id in step_ids:
                    generated = generate_pipeline_endpoint_name(step_id, field_name)
                    result[generated] = generated

    return result


class PipelineStructure(object):
    """An object that holds one or several steps, and describes the connections between them."""

    def __init__(
        self,
        config: "PipelineConfig",
        kiara: typing.Optional["Kiara"] = None,
    ):

        self._structure_config: "PipelineConfig" = config

        steps = self._structure_config.steps
        input_aliases = self._structure_config.input_aliases
        output_aliases = self._structure_config.output_aliases

        if not steps:
            raise Exception("No steps provided.")

        if kiara is None:
            kiara = Kiara.instance()
        self._kiara: Kiara = kiara
        self._steps: typing.List[PipelineStep] = PipelineStep.create_steps(
            *steps, kiara=self._kiara
        )
        # self._pipeline_id: str = parent_id

        if input_aliases is None:
            input_aliases = {}

        if isinstance(input_aliases, str):
            if input_aliases not in ALLOWED_INPUT_ALIAS_MARKERS:
                raise Exception(
                    f"Can't create pipeline, invalid value '{input_aliases}' for 'input_aliases'. Either specify a dict, or use one of: {', '.join(ALLOWED_INPUT_ALIAS_MARKERS)}"
                )

            input_aliases = calculate_shortest_field_aliases(
                self._steps, input_aliases, "inputs"
            )

        if isinstance(output_aliases, str):
            if output_aliases not in ALLOWED_INPUT_ALIAS_MARKERS:
                raise Exception(
                    f"Can't create pipeline, invalid value '{output_aliases}' for 'output_aliases'. Either specify a dict, or use one of: {', '.join(ALLOWED_INPUT_ALIAS_MARKERS)}"
                )

            output_aliases = calculate_shortest_field_aliases(
                self._steps, output_aliases, "outputs"
            )

        self._input_aliases: typing.Dict[str, str] = dict(input_aliases)  # type: ignore
        if output_aliases is None:
            output_aliases = {}
        self._output_aliases: typing.Dict[str, str] = dict(output_aliases)  # type: ignore

        # this is hardcoded for now
        self._add_all_workflow_outputs: bool = False

        self._constants: typing.Dict[str, typing.Any] = None  # type: ignore
        self._defaults: typing.Dict[str, typing.Any] = None  # type: ignore

        self._execution_graph: nx.DiGraph = None  # type: ignore
        self._data_flow_graph: nx.DiGraph = None  # type: ignore
        self._data_flow_graph_simple: nx.DiGraph = None  # type: ignore

        self._processing_stages: typing.List[typing.List[str]] = None  # type: ignore

        self._steps_details: typing.Dict[str, typing.Any] = None  # type: ignore
        """Holds details about the (current) processing steps contained in this workflow."""

    # @property
    # def pipeline_id(self) -> str:
    #     return self._pipeline_id

    @property
    def structure_config(self) -> "PipelineConfig":
        return self._structure_config

    @property
    def steps(self) -> typing.Iterable[PipelineStep]:
        return self._steps

    @property
    def modules(self) -> typing.Iterable["KiaraModule"]:
        return (s.module for s in self.steps)

    @property
    def steps_details(self) -> typing.Mapping[str, typing.Any]:

        if self._steps_details is None:
            self._process_steps()
        return self._steps_details

    @property
    def step_ids(self) -> typing.Iterable[str]:
        if self._steps_details is None:
            self._process_steps()
        return self._steps_details.keys()

    @property
    def constants(self) -> typing.Mapping[str, typing.Any]:

        if self._constants is None:
            self._process_steps()
        return self._constants

    @property
    def defaults(self) -> typing.Mapping[str, typing.Any]:

        if self._defaults is None:
            self._process_steps()
        return self._defaults

    def get_step(self, step_id: str) -> PipelineStep:

        d = self.steps_details.get(step_id, None)
        if d is None:
            raise Exception(f"No module with id: {step_id}")

        return d["step"]

    def get_step_inputs(self, step_id: str) -> typing.Mapping[str, StepInputRef]:

        d = self.steps_details.get(step_id, None)
        if d is None:
            raise Exception(f"No module with id: {step_id}")

        return d["inputs"]

    def get_step_outputs(self, step_id: str) -> typing.Mapping[str, StepOutputRef]:

        d = self.steps_details.get(step_id, None)
        if d is None:
            raise Exception(f"No module with id: {step_id}")

        return d["outputs"]

    def get_step_details(self, step_id: str) -> typing.Mapping[str, typing.Any]:

        d = self.steps_details.get(step_id, None)
        if d is None:
            raise Exception(f"No module with id: {step_id}")

        return d

    @property
    def execution_graph(self) -> nx.DiGraph:
        if self._execution_graph is None:
            self._process_steps()
        return self._execution_graph

    @property
    def data_flow_graph(self) -> nx.DiGraph:
        if self._data_flow_graph is None:
            self._process_steps()
        return self._data_flow_graph

    @property
    def data_flow_graph_simple(self) -> nx.DiGraph:
        if self._data_flow_graph_simple is None:
            self._process_steps()
        return self._data_flow_graph_simple

    @property
    def processing_stages(self) -> typing.List[typing.List[str]]:
        if self._steps_details is None:
            self._process_steps()
        return self._processing_stages

    @lru_cache()
    def _get_node_of_type(self, node_type: str):
        if self._steps_details is None:
            self._process_steps()

        return [
            node
            for node, attr in self._data_flow_graph.nodes(data=True)
            if attr["type"] == node_type
        ]

    @property
    def steps_inputs(self) -> typing.Dict[str, StepInputRef]:
        return {
            node.alias: node
            for node in self._get_node_of_type(node_type=StepInputRef.__name__)
        }

    @property
    def steps_outputs(self) -> typing.Dict[str, StepOutputRef]:
        return {
            node.alias: node
            for node in self._get_node_of_type(node_type=StepOutputRef.__name__)
        }

    @property
    def pipeline_inputs(self) -> typing.Dict[str, PipelineInputRef]:
        return {
            node.value_name: node
            for node in self._get_node_of_type(node_type=PipelineInputRef.__name__)
        }

    @property
    def pipeline_outputs(self) -> typing.Dict[str, PipelineOutputRef]:
        return {
            node.value_name: node
            for node in self._get_node_of_type(node_type=PipelineOutputRef.__name__)
        }

    @property
    def pipeline_input_schema(self) -> typing.Mapping[str, ValueSchema]:

        return {
            input_name: w_in.value_schema
            for input_name, w_in in self.pipeline_inputs.items()
        }

    @property
    def pipeline_output_schema(self) -> typing.Mapping[str, ValueSchema]:
        return {
            output_name: w_out.value_schema
            for output_name, w_out in self.pipeline_outputs.items()
        }

    def _process_steps(self):
        """The core method of this class, it connects all the processing modules, their inputs and outputs."""

        steps_details: typing.Dict[str, typing.Any] = {}
        execution_graph = nx.DiGraph()
        execution_graph.add_node("__root__")
        data_flow_graph = nx.DiGraph()
        data_flow_graph_simple = nx.DiGraph()
        processing_stages = []
        constants = {}
        structure_defaults = {}

        # temp variable, to hold all outputs
        outputs: typing.Dict[str, StepOutputRef] = {}

        # process all pipeline and step outputs first
        _temp_steps_map: typing.Dict[str, PipelineStep] = {}
        pipeline_outputs: typing.Dict[str, PipelineOutputRef] = {}
        for step in self._steps:

            _temp_steps_map[step.step_id] = step

            if step.step_id in steps_details.keys():
                raise Exception(
                    f"Can't process steps: duplicate step_id '{step.step_id}'"
                )

            steps_details[step.step_id] = {
                "step": step,
                "outputs": {},
                "inputs": {},
            }

            data_flow_graph.add_node(step, type="step")

            # go through all the module outputs, create points for them and connect them to pipeline outputs
            for output_name, schema in step.module.output_schemas.items():

                step_output = StepOutputRef(
                    value_name=output_name,
                    value_schema=schema,
                    step_id=step.step_id,
                )

                steps_details[step.step_id]["outputs"][output_name] = step_output
                step_alias = generate_step_alias(step.step_id, output_name)
                outputs[step_alias] = step_output

                step_output_name = generate_pipeline_endpoint_name(
                    step_id=step.step_id, value_name=output_name
                )
                if self._output_aliases:
                    if step_output_name in self._output_aliases.keys():
                        step_output_name = self._output_aliases[step_output_name]
                    else:
                        if not self._add_all_workflow_outputs:
                            # this output is not interesting for the workflow
                            step_output_name = None

                if step_output_name:
                    step_output_address = StepValueAddress(
                        step_id=step.step_id, value_name=output_name
                    )
                    pipeline_output = PipelineOutputRef(
                        value_name=step_output_name,
                        connected_output=step_output_address,
                        value_schema=schema,
                    )
                    pipeline_outputs[step_output_name] = pipeline_output
                    step_output.pipeline_output = pipeline_output.value_name

                    data_flow_graph.add_node(
                        pipeline_output, type=PipelineOutputRef.__name__
                    )
                    data_flow_graph.add_edge(step_output, pipeline_output)

                    data_flow_graph_simple.add_node(
                        pipeline_output, type=PipelineOutputRef.__name__
                    )
                    data_flow_graph_simple.add_edge(step, pipeline_output)

                data_flow_graph.add_node(step_output, type=StepOutputRef.__name__)
                data_flow_graph.add_edge(step, step_output)

        # now process inputs, and connect them to the appropriate output/pipeline-input points
        existing_pipeline_input_points: typing.Dict[str, PipelineInputRef] = {}
        for step in self._steps:

            other_step_dependency: typing.Set = set()
            # go through all the inputs of a module, create input points and connect them to either
            # other module outputs, or pipeline inputs (which need to be created)

            module_constants: typing.Mapping[
                str, typing.Any
            ] = step.module.get_config_value("constants")

            for input_name, schema in step.module.input_schemas.items():

                matching_input_links: typing.List[StepValueAddress] = []
                is_constant = input_name in module_constants.keys()

                for value_name, input_links in step.input_links.items():
                    if value_name == input_name:
                        for input_link in input_links:
                            if input_link in matching_input_links:
                                raise Exception(f"Duplicate input link: {input_link}")
                            matching_input_links.append(input_link)

                if matching_input_links:
                    # this means we connect to other steps output

                    connected_output_points: typing.List[StepOutputRef] = []
                    connected_outputs: typing.List[StepValueAddress] = []

                    for input_link in matching_input_links:
                        output_id = generate_step_alias(
                            input_link.step_id, input_link.value_name
                        )

                        if output_id not in outputs.keys():
                            raise Exception(
                                f"Can't connect input '{input_name}' for step '{step.step_id}': no output '{output_id}' available. Available output names: {', '.join(outputs.keys())}"
                            )
                        connected_output_points.append(outputs[output_id])
                        connected_outputs.append(input_link)

                        other_step_dependency.add(input_link.step_id)

                    step_input_point = StepInputRef(
                        step_id=step.step_id,
                        value_name=input_name,
                        value_schema=schema,
                        is_constant=is_constant,
                        connected_pipeline_input=None,
                        connected_outputs=connected_outputs,
                    )

                    for op in connected_output_points:
                        op.connected_inputs.append(step_input_point.address)
                        data_flow_graph.add_edge(op, step_input_point)
                        data_flow_graph_simple.add_edge(
                            _temp_steps_map[op.step_id], step_input_point
                        )  # TODO: name edge
                        data_flow_graph_simple.add_edge(
                            step_input_point, step
                        )  # TODO: name edge

                else:
                    # this means we connect to pipeline input
                    pipeline_input_name = generate_pipeline_endpoint_name(
                        step_id=step.step_id, value_name=input_name
                    )
                    # check whether this input has an alias associated with it
                    if self._input_aliases:
                        if pipeline_input_name in self._input_aliases.keys():
                            # this means we use the pipeline alias
                            pipeline_input_name = self._input_aliases[
                                pipeline_input_name
                            ]

                    if pipeline_input_name in existing_pipeline_input_points.keys():
                        # we already created a pipeline input with this name
                        # TODO: check whether schema fits
                        connected_pipeline_input = existing_pipeline_input_points[
                            pipeline_input_name
                        ]
                        assert connected_pipeline_input.is_constant == is_constant
                    else:
                        # we need to create the pipeline input
                        connected_pipeline_input = PipelineInputRef(
                            value_name=pipeline_input_name,
                            value_schema=schema,
                            is_constant=is_constant,
                        )

                        existing_pipeline_input_points[
                            pipeline_input_name
                        ] = connected_pipeline_input

                        data_flow_graph.add_node(
                            connected_pipeline_input, type=PipelineInputRef.__name__
                        )
                        data_flow_graph_simple.add_node(
                            connected_pipeline_input, type=PipelineInputRef.__name__
                        )
                        if is_constant:
                            constants[
                                pipeline_input_name
                            ] = step.module.get_config_value("constants")[input_name]

                        default_val = step.module.get_config_value("defaults").get(
                            input_name, None
                        )
                        if is_constant and default_val is not None:
                            raise Exception(
                                f"Module config invalid for step '{step.step_id}': both default value and constant provided for input '{input_name}'."
                            )
                        elif default_val is not None:
                            structure_defaults[pipeline_input_name] = default_val

                    step_input_point = StepInputRef(
                        step_id=step.step_id,
                        value_name=input_name,
                        value_schema=schema,
                        connected_pipeline_input=connected_pipeline_input.value_name,
                        connected_outputs=None,
                    )
                    connected_pipeline_input.connected_inputs.append(
                        step_input_point.address
                    )
                    data_flow_graph.add_edge(connected_pipeline_input, step_input_point)
                    data_flow_graph_simple.add_edge(connected_pipeline_input, step)

                data_flow_graph.add_node(step_input_point, type=StepInputRef.__name__)

                steps_details[step.step_id]["inputs"][input_name] = step_input_point

                data_flow_graph.add_edge(step_input_point, step)

            if other_step_dependency:
                for module_id in other_step_dependency:
                    execution_graph.add_edge(module_id, step.step_id)
            else:
                execution_graph.add_edge("__root__", step.step_id)

        # calculate execution order
        path_lengths: typing.Dict[str, int] = {}

        for step in self._steps:

            step_id = step.step_id

            paths = list(nx.all_simple_paths(execution_graph, "__root__", step_id))
            max_steps = max(paths, key=lambda x: len(x))
            path_lengths[step_id] = len(max_steps) - 1

        max_length = max(path_lengths.values())

        for i in range(1, max_length + 1):
            stage: typing.List[str] = [
                m for m, length in path_lengths.items() if length == i
            ]
            processing_stages.append(stage)
            for _step_id in stage:
                steps_details[_step_id]["processing_stage"] = i
                steps_details[_step_id]["step"].processing_stage = i

        self._constants = constants
        self._defaults = structure_defaults
        self._steps_details = steps_details
        self._execution_graph = execution_graph
        self._data_flow_graph = data_flow_graph
        self._data_flow_graph_simple = data_flow_graph_simple
        self._processing_stages = processing_stages

        self._get_node_of_type.cache_clear()

        # calculating which steps are always required to execute to compute one of the required pipeline outputs.
        # this is done because in some cases it's possible that some steps can be skipped to execute if they
        # don't have a valid input set, because the inputs downstream they are connecting to are 'non-required'
        # optional_steps = []

        last_stage = self._processing_stages[-1]

        step_nodes: typing.List[PipelineStep] = [
            node
            for node in self._data_flow_graph_simple.nodes
            if isinstance(node, PipelineStep)
        ]

        all_required_inputs = []
        for step_id in last_stage:

            step = self.get_step(step_id)
            step_nodes.remove(step)

            for k, s_inp in self.get_step_inputs(step_id).items():
                if not s_inp.value_schema.is_required():
                    continue
                all_required_inputs.append(s_inp)

        for pipeline_input in self.pipeline_inputs.values():

            for last_step_input in all_required_inputs:
                try:
                    path = nx.shortest_path(
                        self._data_flow_graph_simple, pipeline_input, last_step_input
                    )
                    for p in path:
                        if p in step_nodes:
                            step_nodes.remove(p)
                except (NetworkXNoPath, NodeNotFound):
                    pass
                    # print("NO PATH")
                    # print(f"{pipeline_input} -> {last_step_input}")

        for s in step_nodes:
            s.required = False

        for input_name, inp in self.pipeline_inputs.items():
            steps = set()
            for ci in inp.connected_inputs:
                steps.add(ci.step_id)

            optional = True
            for step_id in steps:
                step = self.get_step(step_id)
                if step.required:
                    optional = False
                    break
            if optional:
                inp.value_schema.optional = True

    def extend(
        self,
        other: typing.Union[
            "Pipeline",
            "PipelineStructure",
            "PipelineConfig",
            typing.Mapping[str, typing.Any],
        ],
        input_links: typing.Optional[
            typing.Mapping[str, typing.Iterable[StepValueAddress]]
        ] = None,
    ) -> "PipelineStructure":

        return extend_pipeline(self, other)

    def to_details(self) -> "PipelineStructureDesc":

        from kiara.info.pipelines import PipelineStructureDesc

        return PipelineStructureDesc.create_pipeline_structure_desc(pipeline=self)

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        d = self.to_details()
        yield d



# -*- coding: utf-8 -*-
import typing

from kiara.rendering import KiaraRenderer
from kiara.rendering.jinja import JinjaPipelineRenderer

if typing.TYPE_CHECKING:
    from kiara import Kiara

DEFAULT_TEMPLATES = {"pipeline": "pipeline"}


class TemplateRenderingMgmt(object):
    @classmethod
    def create(
        cls,
        templates: typing.Optional[
            typing.Mapping[str, typing.Union[typing.Mapping[str, typing.Any], str]]
        ] = None,
        kiara: typing.Optional["Kiara"] = None,
    ) -> "TemplateRenderingMgmt":

        if not templates:
            templates = DEFAULT_TEMPLATES

        _templates: typing.Dict[str, typing.Dict[str, typing.Any]] = {}
        for template_name, config in templates.items():
            if isinstance(config, type) and issubclass(config, KiaraRenderer):
                _templates[template_name] = {
                    "renderer_cls": config,
                    "renderer_config": {},
                }
            elif isinstance(config, str):
                if config == "pipeline":
                    _templates[template_name] = {
                        "renderer_cls": JinjaPipelineRenderer,
                        "renderer_config": {},
                    }
                else:
                    raise NotImplementedError()
            else:
                raise NotImplementedError()

        return TemplateRenderingMgmt(templates=_templates, kiara=kiara)

    def __init__(
        self,
        templates: typing.Mapping[str, typing.Mapping[str, typing.Any]],
        kiara: typing.Optional["Kiara"] = None,
    ):

        if kiara is None:
            from kiara import Kiara

            kiara = Kiara.instance()

        self._kiara: Kiara = kiara
        self._templates: typing.Mapping[
            str, typing.Mapping[str, typing.Any]
        ] = templates

    def render(self, template_name: str, **inputs: typing.Any) -> typing.Any:

        if template_name not in self._templates.keys():
            raise Exception(
                f"No template '{template_name}' registered. Available templates: {', '.join(self._templates.keys())}"
            )

        template_cls = self._templates[template_name]["renderer_cls"]
        template_config = self._templates[template_name]["renderer_config"]

        template_renderer: KiaraRenderer = template_cls(
            config=template_config, kiara=self._kiara
        )

        return template_renderer.render(**inputs)



# -*- coding: utf-8 -*-

import os
from click.testing import CliRunner

from kiara.interfaces.cli import cli


def test_pipeline_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "pipeline")
    assert result.exit_code == 0
    assert "Pipeline-related sub-commands" in result.stdout


def test_pipeline_explain_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "pipeline explain logic.xor")

    assert result.exit_code == 0
    assert "Pipeline structure" in result.stdout
    assert "Processing stage: 2" in result.stdout

    pipeline_file = os.path.join(
        os.path.dirname(__file__),
        "..",
        "resources",
        "pipelines",
        "logic",
        "logic_3.json",
    )
    abs_path = os.path.abspath(pipeline_file)
    result = runner.invoke(cli, f"pipeline explain '{abs_path}'")

    assert result.exit_code == 0
    assert "Pipeline structure" in result.stdout
    assert "Processing stage: 2" in result.stdout


def test_pipeline_explain_steps_subcommand():

    runner = CliRunner()
    result = runner.invoke(cli, "pipeline explain-steps logic.xor")

    assert result.exit_code == 0
    assert "logic.or" in result.stdout
    assert "logic.nand" in result.stdout
    assert "logic.and" in result.stdout

    pipeline_file = os.path.join(
        os.path.dirname(__file__),
        "..",
        "resources",
        "pipelines",
        "logic",
        "logic_3.json",
    )
    abs_path = os.path.abspath(pipeline_file)
    result = runner.invoke(cli, f"pipeline explain '{abs_path}'")

    assert result.exit_code == 0
    assert "logic.and" in result.stdout



# -*- coding: utf-8 -*-
import click
import os

from kiara import Kiara
from kiara.defaults import KIARA_RESOURCES_FOLDER


@click.group("dev")
@click.pass_context
def dev_group(ctx):
    """Development helpers."""


@dev_group.command("test")
@click.pass_context
def test(ctx):

    kiara_obj: Kiara = ctx.obj["kiara"]

    pipeline = "/home/markus/projects/dharpa/kiara-playground/examples/streamlit/geolocation_prototype/pipelines/geolocation_1.yml"

    template = os.path.join(KIARA_RESOURCES_FOLDER, "templates", "python_script.py.j2")
    rendered = kiara_obj.template_mgmt.render(
        "pipeline_notebook", module=pipeline, template=template
    )

    print(rendered)

    # inputs = {
    #     "edges_file_path": "/home/markus/projects/dharpa/kiara-playground/examples/data/journals/JournalEdges1902.csv",
    #     "nodes_file_path": "/home/markus/projects/dharpa/kiara-playground/examples/data/journals/JournalNodes1902.csv"
    # }
    # pipeline = "/home/markus/projects/dharpa/kiara.streamlit/apps/component_gallery/pipelines/gallery_onboarding.yaml"
    #
    # store_config_dict = {
    #     "outputs": [
    #         {"alias_template": "{{ base_id }}__{{ field_name }}"}
    #     ]
    # }
    #
    # onboard_config = {
    #     "module_type": pipeline,
    #     "inputs": inputs,
    #     "store_config": store_config_dict
    # }
    #
    # # store_config = ValueStoreConfig(**store_config_dict)
    # onboarder = BatchOnboard.create(kiara=kiara_obj, **onboard_config)
    # results = onboarder.run("test")
    # dbg(results)



# -*- coding: utf-8 -*-

from click.testing import CliRunner

from kiara.interfaces.cli import cli


def test_non_arg_cli():

    runner = CliRunner()
    result = runner.invoke(cli)
    result.exit_code == 0
    assert "Metadata-related sub-commands.\n" in result.stdout

    result_2 = runner.invoke(cli, "--help")
    assert result.stdout == result_2.stdout



# -*- coding: utf-8 -*-
import importlib
import inspect
import typing
from pydantic import AnyUrl, BaseModel, EmailStr, Field
from rich import box
from rich.console import RenderableType
from rich.markdown import Markdown
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table
from types import ModuleType

from kiara.defaults import DEFAULT_NO_DESC_VALUE
from kiara.metadata import MetadataModel
from kiara.utils import merge_dicts
from kiara.utils.global_metadata import get_metadata_for_python_module_or_class

# class HashMetadata(MetadataModel):
#
#     hash: str = Field(description="The hash for the value.")
#     hash_desc: typing.Optional[str] = Field(
#         description="A description how the hash was calculated and other details.",
#         default=None,
#     )


class LinkModel(BaseModel):

    url: AnyUrl = Field(description="The url.")
    desc: typing.Optional[str] = Field(
        description="A short description of the link content.",
        default=DEFAULT_NO_DESC_VALUE,
    )


class AuthorModel(BaseModel):

    name: str = Field(description="The full name of the author.")
    email: typing.Optional[EmailStr] = Field(
        description="The email address of the author", default=None
    )


class ContextMetadataModel(MetadataModel):
    @classmethod
    def from_class(cls, item_cls: typing.Type):

        data = get_metadata_for_python_module_or_class(item_cls)  # type: ignore
        merged = merge_dicts(*data)
        return cls.parse_obj(merged)

    _metadata_key = "properties"

    references: typing.Dict[str, LinkModel] = Field(
        description="References for the item.", default_factory=dict
    )
    tags: typing.Set[str] = Field(
        description="A list of tags for the item.", default_factory=set
    )
    labels: typing.Dict[str, str] = Field(
        description="A list of labels for the item.", default_factory=list
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")

        if self.tags:
            table.add_row("Tags", ", ".join(self.tags))
        if self.labels:
            labels = []
            for k, v in self.labels.items():
                labels.append(f"[i]{k}[/i]: {v}")
            table.add_row("Labels", "\n".join(labels))

        if self.references:
            references = []
            for _k, _v in self.references.items():
                link = f"[link={_v.url}]{_v.url}[/link]"
                references.append(f"[i]{_k}[/i]: {link}")
            table.add_row("References", "\n".join(references))

        return table

    def add_reference(
        self,
        ref_type: str,
        url: str,
        desc: typing.Optional[str] = None,
        force: bool = False,
    ):

        if ref_type in self.references.keys() and not force:
            raise Exception(f"Reference of type '{ref_type}' already present.")
        link = LinkModel(url=url, desc=desc)
        self.references[ref_type] = link

    def get_url_for_reference(self, ref: str) -> typing.Optional[str]:

        link = self.references.get(ref, None)
        if not link:
            return None

        return link.url


class DocumentationMetadataModel(MetadataModel):

    _metadata_key = "documentation"

    @classmethod
    def from_class_doc(cls, item_cls: typing.Type):

        doc = item_cls.__doc__

        if not doc:
            doc = DEFAULT_NO_DESC_VALUE

        doc = inspect.cleandoc(doc)
        return cls.from_string(doc)

    @classmethod
    def from_string(cls, doc: typing.Optional[str]):

        if not doc:
            doc = DEFAULT_NO_DESC_VALUE

        if "\n" in doc:
            desc, doc = doc.split("\n", maxsplit=1)
        else:
            desc = doc
            doc = None

        if doc:
            doc = doc.strip()

        return cls(description=desc.strip(), doc=doc)

    @classmethod
    def from_dict(cls, data: typing.Mapping):

        doc = data.get("doc", None)
        desc = data.get("description", None)
        if desc is None:
            desc = data.get("desc", None)

        if not doc and not desc:
            return cls.from_string(DEFAULT_NO_DESC_VALUE)
        elif doc and not desc:
            return cls.from_string(doc)
        elif desc and not doc:
            return cls.from_string(desc)
        else:
            return DocumentationMetadataModel(description=desc, doc=doc)

    @classmethod
    def create(cls, item: typing.Any):

        if not item:
            return cls.from_string(DEFAULT_NO_DESC_VALUE)
        elif isinstance(item, DocumentationMetadataModel):
            return item
        elif isinstance(item, typing.Mapping):
            return cls.from_dict(item)
        if isinstance(item, type):
            return cls.from_class_doc(item)
        elif isinstance(item, str):
            return cls.from_string(item)
        else:
            raise TypeError(f"Can't create documentation from type '{type(item)}'.")

    description: str = Field(
        description="Short description of the item.", default=DEFAULT_NO_DESC_VALUE
    )
    doc: typing.Optional[str] = Field(
        description="Detailed documentation of the item (in markdown).", default=None
    )

    @property
    def full_doc(self):

        if self.doc:
            return f"{self.description}\n\n{self.doc}"
        else:
            return self.description

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        return Markdown(self.full_doc)


class OriginMetadataModel(MetadataModel):

    _metadata_key = "origin"

    @classmethod
    def from_class(cls, item_cls: typing.Type):

        data = get_metadata_for_python_module_or_class(item_cls)  # type: ignore
        merged = merge_dicts(*data)
        return cls.parse_obj(merged)

    authors: typing.List[AuthorModel] = Field(
        description="The authors/creators of this item.", default_factory=list
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Key", style="i")
        table.add_column("Value")

        authors = []
        for author in self.authors:
            if author.email:
                authors.append(f"{author.name} ({author.email})")
            else:
                authors.append(author.name)
        table.add_row("Authors", "\n".join(authors))

        return table


class PythonClassMetadata(MetadataModel):
    """Python class and module information."""

    _metadata_key: typing.ClassVar[str] = "python_class"

    @classmethod
    def from_class(cls, item_cls: typing.Type):

        conf: typing.Dict[str, typing.Any] = {
            "class_name": item_cls.__name__,
            "module_name": item_cls.__module__,
            "full_name": f"{item_cls.__module__}.{item_cls.__name__}",
        }
        return PythonClassMetadata(**conf)

    class_name: str = Field(description="The name of the Python class.")
    module_name: str = Field(
        description="The name of the Python module this class lives in."
    )
    full_name: str = Field(description="The full class namespace.")

    def get_class(self) -> typing.Type:
        m = self.get_module()
        return getattr(m, self.class_name)

    def get_module(self) -> ModuleType:
        m = importlib.import_module(self.module_name)
        return m


class MetadataModelMetadata(MetadataModel):
    @classmethod
    def from_model_class(cls, model_cls: typing.Type[MetadataModel]):

        origin_md = OriginMetadataModel.from_class(model_cls)
        doc = DocumentationMetadataModel.from_class_doc(model_cls)
        python_class = PythonClassMetadata.from_class(model_cls)
        properties_md = ContextMetadataModel.from_class(model_cls)

        return MetadataModelMetadata(
            type_name=model_cls._metadata_key,  # type: ignore
            documentation=doc,
            origin=origin_md,
            context=properties_md,
            python_class=python_class,
        )

    type_name: str = Field(description="The registered name for this value type.")
    documentation: DocumentationMetadataModel = Field(
        description="Documentation for the value type."
    )
    origin: OriginMetadataModel = Field(
        description="Information about the creator of this value type."
    )
    context: ContextMetadataModel = Field(
        description="Generic properties of this value type."
    )
    python_class: PythonClassMetadata = Field(
        description="The Python class for this value type."
    )

    def create_fields_table(
        self, show_header: bool = True, show_required: bool = True
    ) -> Table:

        type_cls = self.python_class.get_class()

        fields_table = Table(show_header=show_header, box=box.SIMPLE)
        fields_table.add_column("Field name", style="i")
        fields_table.add_column("Type")
        if show_required:
            fields_table.add_column("Required")
        fields_table.add_column("Description")
        for field_name, details in type_cls.__fields__.items():
            field_type = type_cls.schema()["properties"][field_name]["type"]
            info = details.field_info.description
            if show_required:
                req = "yes" if details.required else "no"
                fields_table.add_row(field_name, field_type, req, info)
            else:
                fields_table.add_row(field_name, field_type, info)

        return fields_table

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        include_schema = config.get("display_schema", False)
        include_doc = config.get("include_doc", True)

        table = Table(show_header=False, box=box.SIMPLE)
        table.add_column("Property", style="i")
        table.add_column("Value")

        if include_doc:
            table.add_row(
                "Documentation",
                Panel(self.documentation.create_renderable(), box=box.SIMPLE),
            )
        table.add_row("Origin", self.origin.create_renderable())
        table.add_row("Context", self.context.create_renderable())

        table.add_row("Python class", self.python_class.create_renderable())

        fields_table = self.create_fields_table()
        table.add_row("Fields", fields_table)

        if include_schema:
            json_str = Syntax(
                self.python_class.get_class().schema_json(indent=2),
                "json",
                background_color="default",
            )
            table.add_row("Json Schema", json_str)

        return table



# -*- coding: utf-8 -*-
import threading
import typing

from kiara import Kiara


def test_multiple_kiara_instances():

    kiara = Kiara()
    kiara_2 = Kiara()

    assert kiara.available_operation_ids == kiara_2.available_operation_ids
    assert kiara.available_module_types == kiara_2.available_module_types


def test_multiple_kiara_instances_threaded():
    def create_kiara_context(ops: typing.List[str]):
        kiara = Kiara()
        ops.extend(kiara.available_operation_ids)

    ops_1 = []
    ops_2 = []

    thread_1 = threading.Thread(target=create_kiara_context, args=(ops_1,))
    thread_1.start()

    thread_2 = threading.Thread(target=create_kiara_context, args=(ops_2,))
    thread_2.start()

    thread_1.join()
    thread_2.join()

    assert ops_1
    assert ops_1 == ops_2



# -*- coding: utf-8 -*-
"""Base module for code that handles the import and management of [KiaraModule][kiara.module.KiaraModule] sub-classes."""

import abc
import logging
import typing

if typing.TYPE_CHECKING:
    from kiara import Kiara, KiaraModule
    from kiara.module_config import ModuleTypeConfigSchema
    from kiara.module_mgmt.pipelines import PipelineModuleManagerConfig
    from kiara.module_mgmt.python_classes import (
        PythonModuleManager,
        PythonModuleManagerConfig,
    )


log = logging.getLogger("kiara")


class ModuleManager(abc.ABC):
    @classmethod
    def from_config(
        cls,
        config: typing.Union[
            typing.Mapping[str, typing.Any],
            "PipelineModuleManagerConfig",
            "PythonModuleManagerConfig",
        ],
    ) -> "ModuleManager":

        raise NotImplementedError()

        from kiara.module_mgmt.pipelines import (
            PipelineModuleManager,
            PipelineModuleManagerConfig,
        )
        from kiara.module_mgmt.python_classes import (
            PythonModuleManager,
            PythonModuleManagerConfig,
        )

        if isinstance(config, typing.Mapping):
            mm_type = config.get("module_manager_type", None)
            if not mm_type:
                raise ValueError(f"No module manager type provided in config: {config}")
            if mm_type == "python":
                config = PythonModuleManagerConfig(**config)
            elif mm_type == "pipeline":
                config = PipelineModuleManagerConfig(**config)
            else:
                raise ValueError(f"Invalid module manager type: {mm_type}")

        if config.module_manager_type == "python":
            mm: ModuleManager = PythonModuleManager(
                **config.dict(exclude={"module_manager_type"})
            )
        elif config.module_manager_type == "pipeline":
            mm = PipelineModuleManager(**config.dict(exclude={"module_manager_type"}))

        return mm

    @abc.abstractmethod
    def get_module_types(self) -> typing.Iterable[str]:
        pass

    @abc.abstractmethod
    def get_module_class(self, module_type: str) -> typing.Type["KiaraModule"]:
        pass

    def create_module_config(
        self, module_type: str, module_config: typing.Mapping[str, typing.Any]
    ) -> "ModuleTypeConfigSchema":

        cls = self.get_module_class(module_type)
        config = cls._config_cls(**module_config)

        return config

    def create_module(
        self,
        kiara: "Kiara",
        id: typing.Optional[str],
        module_type: str,
        module_config: typing.Optional[typing.Mapping[str, typing.Any]] = None,
        parent_id: typing.Optional[str] = None,
    ) -> "KiaraModule":

        module_cls = self.get_module_class(module_type)

        module = module_cls(
            id=id, parent_id=parent_id, module_config=module_config, kiara=kiara
        )
        return module


class WorkflowManager(object):
    def __init__(self, module_manager: "PythonModuleManager"):

        self._module_mgr: "PythonModuleManager" = module_manager

    def create_workflow(
        self,
        workflow_id: str,
        config: typing.Union[str, typing.Mapping[str, typing.Any]],
    ):

        if isinstance(config, typing.Mapping):
            raise NotImplementedError()



# -*- coding: utf-8 -*-
import typing
from enum import Enum
from pydantic import BaseModel, Extra, Field

from kiara.data.values import Value, ValueSchema
from kiara.data.values.value_set import ValueSet
from kiara.pipeline.utils import generate_step_alias


class PipelineValueInfo(BaseModel):
    """Convenience wrapper to make the [PipelineState][kiara.pipeline.pipeline.PipelineState] json/dict export prettier."""

    @classmethod
    def from_value_obj(cls, value: Value, ensure_metadata: bool = False):

        if ensure_metadata:
            value.get_metadata()

        return PipelineValueInfo(
            id=value.id,
            value_schema=value.value_schema,
            is_valid=value.item_is_valid(),
            is_set=value.is_set,
            status=value.item_status(),
            # value_metadata=value.value_metadata,
            # last_update=value.last_update,
            # value_hash=value.value_hash,
            # is_streaming=value.is_streaming,
            metadata=value.metadata,
        )

    class Config:
        extra = Extra.forbid
        allow_mutation = False

    id: str = Field(description="A unique id for this value.")
    is_valid: bool = Field(
        description="Whether the value is set and valid.", default=False
    )
    status: str = Field(description="The value status string")
    is_set: bool = Field(description="Whether the value is set.")
    value_schema: ValueSchema = Field(description="The schema of this value.")
    # is_constant: bool = Field(
    #     description="Whether this value is a constant.", default=False
    # )
    # value_metadata: ValueMetadata = Field(
    #     description="The metadata of the value itself (not the actual data)."
    # )
    # last_update: datetime = Field(
    #     default=None, description="The time the last update to this value happened."
    # )
    # value_hash: typing.Union[ValueHashMarker, int] = Field(
    #     description="The hash of the current value."
    # )
    # is_streaming: bool = Field(
    #     default=False,
    #     description="Whether the value is currently streamed into this object.",
    # )
    metadata: typing.Dict[str, typing.Any] = Field(
        description="Metadata relating to the actual data (size, no. of rows, etc. -- depending on data type).",
        default_factory=dict,
    )


class PipelineValuesInfo(BaseModel):
    """Convenience wrapper to make the [PipelineState][kiara.pipeline.pipeline.PipelineState] json/dict export prettier.

    This is basically just a simplified version of the [ValueSet][kiara.data.values.ValueSet] class that is using
    pydantic, in order to make it easy to export to json.
    """

    @classmethod
    def from_value_set(cls, value_set: ValueSet, ensure_metadata: bool = False):

        values: typing.Dict[str, PipelineValueInfo] = {}
        for k in value_set.get_all_field_names():
            v = value_set.get_value_obj(k, ensure_metadata=ensure_metadata)
            values[k] = PipelineValueInfo.from_value_obj(
                v, ensure_metadata=ensure_metadata
            )

        return PipelineValuesInfo(values=values)

    values: typing.Dict[str, PipelineValueInfo] = Field(
        description="Field names are keys, and the data as values."
    )

    class Config:
        use_enum_values = True


class StepValueAddress(BaseModel):
    """Small model to describe the address of a value of a step, within a Pipeline/PipelineStructure."""

    class Config:
        extra = Extra.forbid

    step_id: str = Field(description="The id of a step within a pipeline.")
    value_name: str = Field(
        description="The name of the value (output name or pipeline input name)."
    )
    sub_value: typing.Optional[typing.Dict[str, typing.Any]] = Field(
        default=None,
        description="A reference to a subitem of a value (e.g. column, list item)",
    )

    @property
    def alias(self):
        """An alias string for this address (in the form ``[step_id].[value_name]``)."""
        return generate_step_alias(self.step_id, self.value_name)

    def __eq__(self, other):

        if not isinstance(other, StepValueAddress):
            return False

        return (self.step_id, self.value_name, self.sub_value) == (
            other.step_id,
            other.value_name,
            other.sub_value,
        )

    def __hash__(self):

        return hash((self.step_id, self.value_name, self.sub_value))

    def __repr__(self):

        if self.sub_value:
            sub_value = f" sub_value={self.sub_value}"
        else:
            sub_value = ""
        return f"StepValueAddres(step_id={self.step_id}, value_name={self.value_name}{sub_value})"

    def __str__(self):
        return self.__repr__()


class StepStatus(Enum):
    """Enum to describe the state of a workflow."""

    STALE = "stale"
    INPUTS_READY = "inputs_ready"
    RESULTS_INCOMING = "processing"
    RESULTS_READY = "results_ready"



# -*- coding: utf-8 -*-
from kiara import Kiara


def test_module_types_exist(kiara: Kiara):

    assert "logic.and" in kiara.module_mgmt.get_module_types()
    assert "logic.xor" in kiara.module_mgmt.get_module_types()

    assert "logic.and" not in kiara.module_mgmt.available_pipeline_module_types
    assert "logic.and" in kiara.module_mgmt.available_non_pipeline_module_types

    assert "logic.xor" in kiara.module_mgmt.available_pipeline_module_types
    assert "logic.xor" not in kiara.module_mgmt.available_non_pipeline_module_types


def test_module_type_metadata(kiara: Kiara):

    l_and = kiara.module_mgmt.get_module_class("logic.and")
    assert not l_and.is_pipeline()

    l_xor = kiara.module_mgmt.get_module_class("logic.xor")
    assert l_xor.is_pipeline()

    assert hasattr(l_and, "_module_type_name")
    assert hasattr(l_and, "_module_type_id")

    assert hasattr(l_xor, "_module_type_name")
    assert hasattr(l_xor, "_module_type_id")

    assert (
        l_and.get_type_metadata().documentation.full_doc
        == l_and.get_type_metadata().documentation.description
    )
    assert (
        l_xor.get_type_metadata().documentation.full_doc
        == l_xor.get_type_metadata().documentation.description
    )

    assert l_and.get_type_metadata().context.labels["package"] == "kiara_modules.core"
    assert l_xor.get_type_metadata().context.labels["package"] == "kiara_modules.core"



# -*- coding: utf-8 -*-
import os
import sys
import typing
from appdirs import AppDirs
from enum import Enum

kiara_app_dirs = AppDirs("kiara", "DHARPA")

if not hasattr(sys, "frozen"):
    KIARA_MODULE_BASE_FOLDER = os.path.dirname(__file__)
    """Marker to indicate the base folder for the `kiara` module."""
else:
    KIARA_MODULE_BASE_FOLDER = os.path.join(sys._MEIPASS, "kiara")  # type: ignore
    """Marker to indicate the base folder for the `kiara` module."""

KIARA_RESOURCES_FOLDER = os.path.join(KIARA_MODULE_BASE_FOLDER, "resources")
"""Default resources folder for this package."""

USER_PIPELINES_FOLDER = os.path.join(kiara_app_dirs.user_config_dir, "pipelines")

KIARA_DATA_STORE_DIR = os.path.join(
    kiara_app_dirs.user_data_dir, "data-store", "default"
)
KIARA_DATA_DIR = os.path.join(KIARA_DATA_STORE_DIR, "data")
KIARA_METADATA_DIR = os.path.join(KIARA_DATA_STORE_DIR, "metadata")
KIARA_ALIASES_DIR = os.path.join(KIARA_DATA_STORE_DIR, "aliases")

MODULE_TYPE_KEY = "module_type"
"""The key to specify the type of a module."""

STEP_ID_KEY = "step_id"
"""The key to specify the step id."""

INVALID_VALUE_NAMES = [
    "kiara",
    "registry",
    "dict",
    "items_are_valid",
    "set_values",
    "set_value",
    "ALL",
    "all",
    "metadata",
    "value",
    "value_obj",
    "items",
    "keys",
    "values",
    "data",
    "callbacks",
    "trigger_callbacks",
    "shared_metadata",
]
"""List of reserved names, inputs/outputs can't use those."""

PIPELINE_PARENT_MARKER = "__pipeline__"
"""Marker string in the pipeline structure that indicates a parent pipeline element."""

DEFAULT_EXCLUDE_DIRS = [".git", ".tox", ".cache"]
"""List of directory names to exclude by default when walking a folder recursively."""

DEFAULT_EXCLUDE_FILES = [".DS_Store"]
"""List of file names to exclude by default when reading folders."""

VALID_PIPELINE_FILE_EXTENSIONS = ["yaml", "yml", "json"]
"""File extensions a kiara pipeline/workflow file can have."""

MODULE_TYPE_NAME_KEY = "module_type_name"
"""The string for the module type name in a module configuration dict."""

DEFAULT_PIPELINE_PARENT_ID = "__kiara__"
"""Default parent id for pipeline objects that are not associated with a workflow."""

DEFAULT_NO_DESC_VALUE = "-- n/a --"

KIARA_MODULE_METADATA_ATTRIBUTE = "KIARA_METADATA"


class SpecialValue(Enum):

    NOT_SET = "__not_set__"
    NO_VALUE = "__no_value__"
    IGNORE = "__ignore__"


DEFAULT_PRETTY_PRINT_CONFIG = {
    "max_no_rows": 32,
    "max_row_height": 2,
    "max_cell_length": 80,
}

NO_HASH_MARKER = "--no-hash--"
"""Marker string to indicate no hash was calculated."""

NO_VALUE_ID_MARKER = "--no-value-id--"
"""Marker string to indicate no value id exists."""
DEFAULT_TO_JSON_CONFIG: typing.Mapping[str, typing.Any] = {
    "indent": 2,
}

COLOR_LIST = [
    "green",
    "blue",
    "bright_magenta",
    "dark_red",
    "gold3",
    "cyan",
    "orange1",
    "light_yellow3",
    "light_slate_grey",
    "deep_pink4",
]



# -*- coding: utf-8 -*-
import inspect
import json
import textwrap
import typing
from pydantic import Field, validator
from pydantic.schema import (
    get_flat_models_from_model,
    get_model_name_map,
    model_process_schema,
)
from rich import box
from rich.console import RenderableType
from rich.panel import Panel
from rich.syntax import Syntax
from rich.table import Table

from kiara.data.values import ValueSchema
from kiara.defaults import DEFAULT_NO_DESC_VALUE
from kiara.metadata import MetadataModel, ValueTypeAndDescription
from kiara.metadata.core_models import (
    ContextMetadataModel,
    DocumentationMetadataModel,
    OriginMetadataModel,
    PythonClassMetadata,
)
from kiara.module_config import ModuleTypeConfigSchema
from kiara.pipeline.config import PipelineConfig
from kiara.utils import create_table_from_field_schemas
from kiara.utils.output import create_table_from_base_model

if typing.TYPE_CHECKING:
    from kiara import KiaraModule


class KiaraModuleConfigMetadata(MetadataModel):
    @classmethod
    def from_config_class(
        cls,
        config_cls: typing.Type[ModuleTypeConfigSchema],
        remove_pipeline_config: bool = False,
    ):

        flat_models = get_flat_models_from_model(config_cls)
        model_name_map = get_model_name_map(flat_models)
        m_schema, _, _ = model_process_schema(config_cls, model_name_map=model_name_map)
        fields = m_schema["properties"]

        config_values = {}
        for field_name, details in fields.items():
            if remove_pipeline_config and field_name in [
                "steps",
                "input_aliases",
                "output_aliases",
                "doc",
            ]:
                continue

            type_str = "-- n/a --"
            if "type" in details.keys():
                type_str = details["type"]

            desc = details.get("description", DEFAULT_NO_DESC_VALUE)
            default = config_cls.__fields__[field_name].default
            default = config_cls.__fields__[field_name].default
            if default is None:
                if callable(config_cls.__fields__[field_name].default_factory):
                    default = config_cls.__fields__[field_name].default_factory()  # type: ignore

            req = config_cls.__fields__[field_name].required

            config_values[field_name] = ValueTypeAndDescription(
                description=desc, type=type_str, value_default=default, required=req
            )

        python_cls = PythonClassMetadata.from_class(config_cls)
        return KiaraModuleConfigMetadata(
            python_class=python_cls, config_values=config_values
        )

    python_class: PythonClassMetadata = Field(
        description="The Python class for this configuration."
    )
    config_values: typing.Dict[str, ValueTypeAndDescription] = Field(
        description="The available configuration values."
    )


def calculate_class_doc_url(base_url: str, module_type_name: str, pipeline: bool):

    if base_url.endswith("/"):
        base_url = base_url[0:-1]

    module_type_name = module_type_name.replace(".", "")
    if not pipeline:
        url = f"{base_url}/latest/modules_list/#{module_type_name}"
    else:
        url = f"{base_url}/latest/pipelines_list/#{module_type_name}"

    return url


def calculate_class_source_url(
    base_url: str, python_class_info: PythonClassMetadata, branch: str = "main"
):

    if base_url.endswith("/"):
        base_url = base_url[0:-1]

    m = python_class_info.get_module()
    m_file = m.__file__

    base_url = f"{base_url}/blob/{branch}/src/{python_class_info.module_name.replace('.', '/')}"

    if m_file.endswith("__init__.py"):
        url = f"{base_url}/__init__.py"
    else:
        url = f"{base_url}.py"

    return url


class KiaraModuleTypeMetadata(MetadataModel):
    @classmethod
    def from_module_class(cls, module_cls: typing.Type["KiaraModule"]):

        module_attrs = cls.extract_module_attributes(module_cls=module_cls)
        return cls(**module_attrs)

    @classmethod
    def extract_module_attributes(
        self, module_cls: typing.Type["KiaraModule"]
    ) -> typing.Dict[str, typing.Any]:

        if not hasattr(module_cls, "process"):
            raise Exception(f"Module class '{module_cls}' misses 'process' method.")
        proc_src = textwrap.dedent(inspect.getsource(module_cls.process))  # type: ignore

        origin_md = OriginMetadataModel.from_class(module_cls)
        doc = DocumentationMetadataModel.from_class_doc(module_cls)
        python_class = PythonClassMetadata.from_class(module_cls)
        properties_md = ContextMetadataModel.from_class(module_cls)

        is_pipeline = module_cls.is_pipeline()
        doc_url = properties_md.get_url_for_reference("documentation")
        if doc_url:
            class_doc = calculate_class_doc_url(doc_url, module_cls._module_type_id, pipeline=is_pipeline)  # type: ignore
            properties_md.add_reference(
                "module_doc",
                class_doc,
                "A link to the published, auto-generated module documentation.",
            )

        if not is_pipeline:
            repo_url = properties_md.get_url_for_reference("source_repo")
            if repo_url is not None:
                src_url = calculate_class_source_url(repo_url, python_class)
                properties_md.add_reference(
                    "source_url",
                    src_url,
                    "A link to the published source file that contains this module.",
                )

        config = KiaraModuleConfigMetadata.from_config_class(module_cls._config_cls)
        pipeline_config = None
        if module_cls._module_type_id != "pipeline" and is_pipeline:  # type: ignore
            pipeline_config = module_cls._base_pipeline_config  # type: ignore

        return {
            "type_name": module_cls._module_type_name,  # type: ignore
            "type_id": module_cls._module_type_id,  # type: ignore
            "documentation": doc,
            "origin": origin_md,
            "context": properties_md,
            "python_class": python_class,
            "config": config,
            "is_pipeline": is_pipeline,
            "pipeline_config": pipeline_config,
            "process_src": proc_src,
        }

    type_name: str = Field(description="The registered name for this module type.")
    type_id: str = Field(description="The full type id.")
    documentation: DocumentationMetadataModel = Field(
        description="Documentation for the module."
    )
    origin: OriginMetadataModel = Field(
        description="Information about authorship for the module type."
    )
    context: ContextMetadataModel = Field(
        description="Generic properties of this module (description, tags, labels, references, ...)."
    )
    python_class: PythonClassMetadata = Field(
        description="Information about the Python class for this module type."
    )
    config: KiaraModuleConfigMetadata = Field(
        description="Details on how this module type can be configured."
    )
    is_pipeline: bool = Field(
        description="Whether the module type is a pipeline, or a core module."
    )
    pipeline_config: typing.Optional[PipelineConfig] = Field(
        description="If this module is a pipeline, this field contains the pipeline configuration.",
        default_factory=None,
    )
    process_src: str = Field(
        description="The source code of the process method of the module."
    )

    @validator("documentation", pre=True)
    def validate_doc(cls, value):

        return DocumentationMetadataModel.create(value)

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        include_config_schema = config.get("include_config_schema", True)
        include_src = config.get("include_src", True)
        include_doc = config.get("include_doc", True)

        table = Table(box=box.SIMPLE, show_header=False, padding=(0, 0, 0, 0))
        table.add_column("property", style="i")
        table.add_column("value")

        if include_doc:
            table.add_row(
                "Documentation",
                Panel(self.documentation.create_renderable(), box=box.SIMPLE),
            )
        table.add_row("Origin", self.origin.create_renderable())
        table.add_row("Context", self.context.create_renderable())

        if include_config_schema:
            if self.config:
                config_cls = self.config.python_class.get_class()
                table.add_row("Module config", create_table_from_base_model(config_cls))
            table.add_row("Module config", "-- no config --")

        table.add_row("Python class", self.python_class.create_renderable())

        if include_src:
            if self.is_pipeline:
                json_str = self.pipeline_config.json(indent=2)  # type: ignore
                _config: Syntax = Syntax(json_str, "json", background_color="default")
                table.add_row("Pipeline config", Panel(_config, box=box.HORIZONTALS))
            else:
                _config = Syntax(self.process_src, "python", background_color="default")
                table.add_row(
                    "Processing source code", Panel(_config, box=box.HORIZONTALS)
                )

        return table


class KiaraModuleInstanceMetadata(MetadataModel):
    @classmethod
    def from_module_obj(cls, obj: "KiaraModule"):

        config = obj.config.dict()
        for x in [
            "steps",
            "input_aliases",
            "output_aliases",
            "module_type_name",
            "doc",
            "metadata",
        ]:
            config.pop(x, None)

        type_metadata = KiaraModuleTypeMetadata.from_module_class(obj.__class__)
        result = KiaraModuleInstanceMetadata(
            type_metadata=type_metadata,
            config=config,
            inputs_schema=obj.input_schemas,
            outputs_schema=obj.output_schemas,
        )
        return result

    type_metadata: KiaraModuleTypeMetadata = Field(
        description="Metadata for the module type of this instance."
    )
    config: typing.Dict[str, typing.Any] = Field(
        description="Configuration that was used to create this module instance."
    )
    inputs_schema: typing.Dict[str, ValueSchema] = Field(
        description="The schema for the module inputs."
    )
    outputs_schema: typing.Dict[str, ValueSchema] = Field(
        description="The schema for the module outputs."
    )

    def create_renderable(self, **config: typing.Any) -> RenderableType:

        include_desc = config.get("include_doc", True)

        table = Table(box=box.SIMPLE, show_header=False, show_lines=True)
        table.add_column("Property", style="i")
        table.add_column("Value")

        if include_desc:
            table.add_row("Description", self.type_metadata.documentation.description)

        table.add_row("Origin", self.type_metadata.origin.create_renderable())
        table.add_row("Type context", self.type_metadata.context.create_renderable())
        table.add_row(
            "Python class", self.type_metadata.python_class.create_renderable()
        )
        conf = Syntax(
            json.dumps(self.config, indent=2), "json", background_color="default"
        )
        table.add_row("Configuration", conf)

        constants = self.config.get("constants")
        inputs_table = create_table_from_field_schemas(
            _add_required=True,
            _add_default=True,
            _show_header=True,
            _constants=constants,
            **self.inputs_schema,
        )
        table.add_row("Inputs", inputs_table)
        outputs_table = create_table_from_field_schemas(
            _add_required=False,
            _add_default=False,
            _show_header=True,
            _constants=None,
            **self.outputs_schema,
        )
        table.add_row("Outputs", outputs_table)
        # table.add_row("Source code", self.type_metadata.process_src)

        return table
